from piscat.Trajectory.data_handeling import protein_trajectories_list2dic
import scipy.io
import h5py
import json
import time
import os
import numpy as np
import zipfile
import wget
import ipywidgets as widgets
from ipywidgets import interact, interactive, fixed, interact_manual
from IPython.display import display


def save_mat(data, path, name=''):
    """
    This function saves the array as matlab format.

    Parameters
    ----------
    data: list
        List or array.

    path: str
        Path of the directory that data saves on it.

    name: str
        Name of the save file.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    name = name + '_' + timestr + '.mat'
    filepath = os.path.join(path, name)
    scipy.io.savemat(filepath, {'data': data}, do_compression=True)


def read_mat(path, name=''):
    """
    This function reads the array with matlab format.

    Parameters
    ----------
    path: str
       Path of the directory that data reads from it.

    name: str
       Name of the file.
    """
    filepath = os.path.join(path, name)
    particles = scipy.io.loadmat(filepath)

    if particles['data'].shape[1] != 1:
        p_ = particles['data']
    else:
        p_ = particles['data']
    return p_


def save_dic_to_hdf5(dic_data, path, name):
    """
    This function writes the dictionary data as hdf5 format.

    Parameters
    ----------
    data: dic
        Dictionary data.

    path: str
        Path of the directory that data saves on it.

    name: str
        Name of the save file
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    name = name + '_' + timestr + '.h5'
    filepath = os.path.join(path, name)
    with h5py.File(filepath, "w") as h5file:
        recursively_save_dict_contents_to_group(h5file, '/', dic_data)


def save_list_to_hdf5(list_data, path, name):
    """
    This function writes the list data as hdf5 format.

    Parameters
    ----------
    data: list
       List data.

    path: str
       Path of the directory that data saves on it.

    name: str
       Name of the save file.
    """
    dic_data = protein_trajectories_list2dic(list_data)
    timestr = time.strftime("%Y%m%d-%H%M%S")
    name = name + '_' + timestr + '.h5'
    filepath = os.path.join(path, name)
    with h5py.File(filepath, "w") as h5file:
        recursively_save_dict_contents_to_group(h5file, '/', dic_data)


def recursively_save_dict_contents_to_group(h5file, path, dic):
    for key, item in dic.items():
        if isinstance(item, (np.ndarray, np.int64, np.float64, str, bytes, list)):
            h5file[path + key] = item
        elif isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file, path + key + '/', item)
        else:
            raise ValueError('Cannot save %s bin_type' % type(item))


def load_dict_from_hdf5(filename):
    """
    This function reads the hdf5 file and convert it to dictionary.

    Parameters
    ----------
    filename: str
       Path and name of the hdf5 file.
    """

    with h5py.File(filename, 'r') as h5file:
        return recursively_load_dict_contents_from_group(h5file, None)


def recursively_load_dict_contents_from_group(data_hdf5, key_group=None):

    ans = {}

    if key_group is None:
        dic_keys = list(data_hdf5.keys())
    else:
        dic_keys = key_group

    for key in dic_keys:
        item = data_hdf5[key]
        if isinstance(item, h5py._hl.dataset.Dataset):
            ans[key] = np.asarray(item)
        elif isinstance(item, h5py._hl.group.Group):
            dic_keys_ = list(item.keys())
            ans[key] = recursively_load_dict_contents_from_group(item, dic_keys_)

    return ans


def save_df2csv(df, path, name=''):
    """
    This function writes the panda data frame to CSV file

    Parameters
    ----------
    data: data frame
        Panda data frame.

    path: str
        Path of the directory that data save on it.

    name: str
        Name of the save file
   """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    name = name + '_' + timestr + '.csv'
    filepath = os.path.join(path, name)
    df.to_csv(filepath)


def save_dic2json(data_dictionary, path, name=''):
    """
    This function writes the dictionary data to JSON file

    Parameters
    ----------
    data: dic
        Dictionary data.

    path: str
        Path of the directory that data save on it.

    name: str
        Name of the save file.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    name = name + '_' + timestr + '.json'
    filepath = os.path.join(path, name)
    with open(filepath, 'w') as file:
        file.write(json.dumps(data_dictionary))


def read_json2dic(path, name=''):
    """
    This function reads the JSON file and converts it to dictionary.

    Parameters
    ----------
    path: str
        Path of the directory that data load from it.
        
    name: str
        Name of the JSON file.
    """
    filepath = os.path.join(path, name)
    if os.path.exists(filepath):
        with open(filepath) as json_file:
            history_setting = json.load(json_file)
        return history_setting


def download_url(url, save_path):
    extension = ".zip"
    # change directory from working dir to dir with files
    os.chdir(save_path)
    filename = wget.download(url)
    print('\nStart unzip files --->', end='')
    for item in os.listdir(save_path):  # loop through items in dir
        if item.endswith(extension):  # check for ".zip" extension
            file_name = os.path.abspath(item)  # get full path of files
            zip_ref = zipfile.ZipFile(file_name)  # create zipfile object
            zip_ref.extractall(save_path)  # extract file to dir
            zip_ref.close()  # close file
            os.remove(file_name)  # delete zipped file
    print('Done')


class download_tutorial_data():

    def __init__(self, tutorial_id, flag_status='JPY'):
        self.flag_status = flag_status
        current_path = os.path.abspath(os.path.join('..'))
        save_path = os.path.dirname(current_path)
        save_path = os.path.join(save_path, 'Tutorials')

        try:
            os.mkdir(save_path)
            print("\nDirectory ", save_path, " Created ")
        except FileExistsError:
            print("\nDirectory ", save_path, " already exists")

        try:
            name_mkdir_1 = 'Demo data'
            dr_mk = os.path.join(save_path, name_mkdir_1)
            os.mkdir(dr_mk)
            print("\nThe directory with the name ", name_mkdir_1, " is created in the following path:", save_path)
        except FileExistsError:
            print("\nThe directory with the name ", name_mkdir_1, " already exists in the following path:", save_path)

        if tutorial_id == 'control_video':
            name_mkdir_1 = 'Demo data'
            name_mkdir_2 = 'Control'
            dr_ = os.path.join(save_path, name_mkdir_1, name_mkdir_2)
            if os.path.isdir(dr_):
                print("\nThe data file named ", name_mkdir_2, " already exists in the following path:", os.path.join(save_path, name_mkdir_1))
            else:
                dr_ = os.path.join(save_path, name_mkdir_1)
                self.run_download(url='https://owncloud.gwdg.de/index.php/s/XgYOcvABZs9cD2Q/download', save_path=dr_)

        elif tutorial_id == 'Tutorial3_video':
            name_mkdir_1 = 'Demo data'
            name_mkdir_2 = 'Tutorial3'
            name_mkdir_3 = 'Histogram'

            dr_ = os.path.join(save_path, name_mkdir_1, name_mkdir_2)
            if os.path.isdir(dr_):
                print("Directory ", name_mkdir_2, " already exists!")
            else:
                dr_ = os.path.join(save_path, name_mkdir_1)
                self.run_download(url='https://owncloud.gwdg.de/index.php/s/p2WVZq7gtqOe6ZY/download', save_path=dr_)

            dr_mk = os.path.join(save_path, name_mkdir_1, name_mkdir_3)

            try:
                os.mkdir(dr_mk)
                print("\nDirectory ", name_mkdir_3, " Created ")
            except FileExistsError:
                print("\nDirectory ", name_mkdir_3, " already exists")

    def run_download(self, url, save_path):
        self.url = url
        self.save_path = save_path

        if "JPY_PARENT_PID" in os.environ and self.flag_status == 'JPY':
            self.button = widgets.Button(description='Download', disabled=False, icon='fa-cloud-download')
            self.out = widgets.Output()

            self.button.on_click(self.on_button_clicked)
            display(self.button)
        elif self.flag_status == 'HTML':
            download_url(self.url, self.save_path)

    def on_button_clicked(self, _):
        download_url(self.url, self.save_path)










from __future__ import print_function

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Localization import data_handeling

from tqdm.autonotebook import tqdm
from PySide6.QtCore import *

import trackpy as tp
import pandas as pd

class WorkerSignals(QObject):
    updateProgress = Signal(int)
    result = Signal(object)
    finished = Signal()


class Linking():


    def __init__(self):
        """
        To obtain the temporal activity of each iPSF, we use the Trackpy packages' algorithm.

       References
       ----------
       [1] http://soft-matter.github.io/trackpy/v0.4.2/
       """

        self.cpu = CPUConfigurations()

    def create_link(self, psf_position, search_range=50, memory=1):
        """
        Each iPSF temporal activity is obtained.

        Parameters
        ----------
        psf_position: pandas data frame
            The data frame contains PSFs locations( x, y, frame, sigma, ...)

        search_range: float or tuple
            The maximum distance features can move between frames, optionally per dimension.

        memory: int
            The maximum number of frames during which a feature can vanish, then reappear nearby, and be considered the same particle. 0 by default.

        Returns
        -------
        df_PSF: pandas dataframe
            To the input data frame, append the 'particle' ID column. ( x, y, frame, sigma, particle, ...).

        """
        df_PSF = tp.link_df(psf_position, search_range=search_range, memory=memory)

        return df_PSF

    def sorting_linking(self, df_PSFs):
        """
        This function uses trajectory lengths to sort particles in a dataframe.

        Parameters
        ----------
        df_PSFs: pandas dataframe
            The data frame contains PSFs locations(x, y, frame, sigma, particle, ...)

        Returns
        -------
        total_sort_df_PSFs: pandas dataframe
            The sort version of data frame contains PSFs locations(x, y, frame, sigma, particle, ...)

        """
        his_all_particles = df_PSFs['particle'].value_counts()
        index_particles = his_all_particles.index
        total_sort_df_PSFs = None

        for index_ in tqdm(index_particles):
            sort_df_PSFs = df_PSFs.loc[df_PSFs['particle'] == index_]
            if total_sort_df_PSFs is None:
                total_sort_df_PSFs = sort_df_PSFs
            else:
                total_sort_df_PSFs = total_sort_df_PSFs.append(sort_df_PSFs)

        return total_sort_df_PSFs

    def trajectory_counter(self, df_PSFs):
        """
        This function counts the number of unique particles in the data frame.

        Parameters
        ----------
        df_PSFs: pandas dataframe
            The data frame contains PSFs locations(x, y, frame, sigma, particle,...)

        Returns
        -------
        unique_list: int
            Returns the number of particles in data frame
        """
        if df_PSFs.shape[0] != 0:
            particles_label = df_PSFs['particle'].tolist()
            unique_list = set(particles_label)
            return len(unique_list)
        else:
            return 0




import inspect
import piscat.Analysis
import piscat.BackgroundCorrection
import piscat.InputOutput
import piscat.Localization
import piscat.Preproccessing
import piscat.Trajectory
import piscat.Visualization
import piscat.GUI


def test_download_data():
    from piscat.InputOutput.read_write_data import download_url
    import os
    current_path = os.path.abspath(os.path.join('.'))
    dr_name = os.path.join(current_path, 'TestData')
    try:
        os.mkdir(dr_name)
        print("\nDirectory TestData Created ")
        print("\nStrats to download Test Data ")
        download_url(url='https://owncloud.gwdg.de/index.php/s/v0TJAYN52vtuXk0/download', save_path=current_path)
        print("\nTestData downloaded successfully")
    except FileExistsError:
        print("\nDirectory TestData already exists in package directory. If you are not sure Test "
              "Data download correctly please delete this directory")


def test_can_import_piscat_subpackages():

    assert inspect.isclass(piscat.Analysis.PlotProteinHistogram)
    assert inspect.isclass(piscat.Analysis.ReadProteinAnalysis)
    assert inspect.isfunction(piscat.Analysis.protein_analysis)

    assert inspect.isclass(piscat.BackgroundCorrection.DifferentialRollingAverage)
    assert inspect.isclass(piscat.BackgroundCorrection.NoiseFloor)

    assert inspect.isclass(piscat.InputOutput.download_tutorial_data)
    assert inspect.isclass(piscat.InputOutput.CPUConfigurations)
    assert inspect.isclass(piscat.InputOutput.CameraParameters)
    assert inspect.isclass(piscat.InputOutput.DirectoryType)
    assert inspect.isclass(piscat.InputOutput.StatusLine)
    assert inspect.isclass(piscat.InputOutput.Image2Video)

    assert inspect.isclass(piscat.Localization.DirectionalIntensity)
    assert inspect.isclass(piscat.Localization.PSFsExtraction)
    assert inspect.isclass(piscat.Localization.SpatialFilter)
    assert inspect.isclass(piscat.Localization.RadialCenter)
    assert inspect.isclass(piscat.Localization.CPUConfigurations)

    assert inspect.isclass(piscat.Preproccessing.ImagePatching)
    assert inspect.isclass(piscat.Preproccessing.Normalization)
    assert inspect.isclass(piscat.Preproccessing.Filters)
    assert inspect.isclass(piscat.Preproccessing.FFT2D)
    assert inspect.isclass(piscat.Preproccessing.FastRadialSymmetryTransform)
    assert inspect.isclass(piscat.Preproccessing.RadialVarianceTransform)

    assert inspect.isclass(piscat.Trajectory.TemporalFilter)
    assert inspect.isclass(piscat.Trajectory.Linking)
    assert inspect.isfunction(piscat.Trajectory.protein_trajectories_list2dic)

    assert inspect.isclass(piscat.Visualization.Display)
    assert inspect.isclass(piscat.Visualization.JupyterDisplay)
    assert inspect.isclass(piscat.Visualization.JupyterPSFs_localizationPreviewDisplay)
    assert inspect.isclass(piscat.Visualization.DisplayPSFs_subplotLocalizationDisplay)
    assert inspect.isclass(piscat.Visualization.JupyterSelectedPSFs_localizationDisplay)
    assert inspect.isclass(piscat.Visualization.ContrastAdjustment)
    assert inspect.isclass(piscat.Visualization.PrintColors)





import numpy as np
import pandas as pd


def feature2df(feature_position, videos):

    if feature_position.shape[1] == 4:

        sigma = feature_position[:, 3]
        psf_position_x = feature_position[:, 2]
        psf_position_y = feature_position[:, 1]
        psf_position_frame = np.asarray(feature_position[:, 0], dtype=int)

        center_intensity = np.asarray([videos[int(psf_position_frame[i_]), int(psf_position_y[i_]),
                                              int(psf_position_x[i_])] for i_ in range(psf_position_frame.shape[0])])

        dict = {'y': psf_position_y, 'x': psf_position_x, 'frame': psf_position_frame, 'center_intensity': center_intensity, 'sigma': sigma}

        df_features = pd.DataFrame(dict)

        return df_features

    elif feature_position.shape[1] == 5:
        sigma = (feature_position[:, 3], feature_position[:, 4])
        sigma = np.asarray(sigma)
        psf_position_x = feature_position[:, 2]
        psf_position_y = feature_position[:, 1]
        psf_position_frame = np.asarray(feature_position[:, 0], dtype=np.int)

        center_intensity = np.asarray([videos[int(psf_position_frame[i_]), int(
            psf_position_y[i_]), int(psf_position_x[i_])] for i_ in
                                       range(psf_position_frame.shape[0])])

        dict = {'y': psf_position_y, 'x': psf_position_x, 'frame': psf_position_frame, 'center_intensity': center_intensity,
                'sigma_x': feature_position[:, 3], 'sigma_y': feature_position[:, 4]}

        df_features = pd.DataFrame(dict)
        df_features['sigma'] = 0.5 * (df_features['sigma_x'] + df_features['sigma_y'])
        return df_features


def list2dataframe(feature_position, video):
    """
    This function converts the output of ``particle_localization.PSFsExtraction`` method from list to data frame.

    Parameters
    ----------
    feature_position: list
        List of position of PSFs (x, y, frame, sigma)

    video: NDArray
        The video is 3D-numpy (number of frames, width, height).

    Returns
    -------
    df_features: pandas dataframe
        PSF positions are stored in the data frame. ( 'y', 'x', 'frame', 'center_intensity', 'sigma', 'Sigma_ratio', ...).
    """
    if feature_position is not None:
        if type(feature_position) is list:
            try:
                feature_position = [l_ for l_ in feature_position if l_ is not None]
                if len(feature_position) >= 1:
                    feature_position = np.concatenate(feature_position, axis=0)
                    df_features = feature2df(feature_position, video)
                else:
                    df_features = None
            except:
                raise Exception('---List is empty!---')

        elif type(feature_position) is np.ndarray:
            try:
                df_features = feature2df(feature_position, video)
            except:
                raise Exception('---List is empty!---')
        else:
            df_features = None

        return df_features
    else:
        df_features = None

        return df_features





from __future__ import print_function
from skimage import io
from PIL import Image
from skimage.color import rgb2gray
from astropy.io import fits
from tqdm.autonotebook import tqdm
from joblib import Parallel, delayed
import os
import pandas as pd
import numpy as np
import cv2
import tifffile
from flifile import FliFile


def video_reader(file_name, type='binary', img_width=128, img_height=128, image_type=np.dtype('<f8'), s_frame=0, e_frame=-1):
    """
    This is a wrapper that can be used to call various video/image readers.

    Parameters
    ----------
    file_name: str
        Path of video and file name, e.g. test.jpg.

    type: str
        Define the video/image format to be loaded.

            * 'binary': use this flag to load binary
            * 'tif': use this flag to load tif
            * 'avi': use this flag to load avi
            * 'png': use this flag to load png
            * 'fits': use this flag to load fits
            * 'fli': use this flag to load fli


    optional_parameters:
        These parameters are used when video 'bin_type' define as binary.

        img_width: int
             For binary images, it specifies the image width.

        img_height: int
            For binary images, it specifies the image height.

        image_type: str
            Numpy.dtype('<u2') --> video with uint16 pixels data type

            * "i"  (signed) integer, "u" unsigned integer, "f" floating-point
            * "<" active little-endian
            * "1" 8-bit, "2" 16-bit, "4" 32-bit, "8" 64-bit

        s_frame: int
            Video reads from this frame. This is used for cropping a video.

        e_frame: int
            Video reads until this frame. This is used for cropping a video.

    Returns
    -------
    @returns: NDArray
        The video/image
    """
    if type == 'binary':
        video = read_binary(file_name, img_width=img_width, img_height=img_height, image_type=image_type, s_frame=s_frame, e_frame=e_frame)
    elif type == 'tif':
        video = read_tif(file_name)
    elif type == 'avi':
        video = read_avi(file_name)
    elif type == 'png':
        video = read_png(file_name)
    elif type == 'fits':
        video = read_fits(file_name)
    elif type == 'fli':
        video = read_fli(file_name)
    return video


def read_binary(file_name, img_width=128, img_height=128, image_type=np.dtype('<f8'), s_frame=0, e_frame=-1):
    """
    This function reads binary video.

    Parameters
    ----------
    file_name: str
        Path and name of binary video.

    img_width: int
         It specifies the image width.

    img_height: int
        It specifies the image height.

    image_type: str

        * "i"  (signed) integer, "u" unsigned integer, "f" floating-point
        * "<" active little-endian
        * "1" 8-bit, "2" 16-bit, "4" 32-bit, "8" 64-bit

    s_frame: int
        Video reads from this frame. This is used for cropping a video.

    e_frame: int
        Video reads until this frame. This is used for cropping a video.

    Returns
    -------
    @returns: NDArray
        The video is 3D-numpy (number of frames, width, height).
    """
    if e_frame == -1:
        num_selected_frames = -1
    else:
        num_selected_frames = (e_frame - s_frame) * (img_width * img_height)
    offset = s_frame * (img_width * img_height)

    img_bin = np.fromfile(file_name, image_type, count=num_selected_frames, sep='', offset=offset)
    number_of_frame = int(img_bin.size / (img_width * img_height))
    return np.reshape(img_bin, (number_of_frame, img_width, img_height))


def read_tif(filename):
    """
    Reading image/video with TIF format.

    Parameters
    ----------
    file_name: str
        Path and name of TIF image/video.

    Returns
    -------
    @returns: NDArray
        The video is 3D-numpy (number of frames, width, height).

    """
    vid_ = tifffile.imread(filename)

    if vid_.ndim == 4:
        frames_list = []
        for i_ in range(vid_.shape[0]):
            gray = cv2.cvtColor(vid_[i_, ...], cv2.COLOR_BGR2GRAY)
            frames_list.append(gray)
    else:
        frames_list = vid_

    frames_list = np.asarray(frames_list)
    return frames_list


def read_tif_iterate(filename):
    """
    Reading image/video with TIF format.

    Parameters
    ----------
    file_name: str
        Path and name of TIF image/video.

    Returns
    -------
    @returns: NDArray
        The video is 3D-numpy (number of frames, width, height).

    """
    vid_ = []
    with tifffile.TiffFile(filename) as tif:
        for page in tqdm(tif.pages):
            image = page.asarray()
            vid_.append(image)
    vid_ = np.array(vid_)

    if vid_.ndim == 4:
        frames_list = []
        for i_ in range(vid_.shape[0]):
            gray = cv2.cvtColor(vid_[i_, ...], cv2.COLOR_BGR2GRAY)
            frames_list.append(gray)
    else:
        frames_list = vid_

    frames_list = np.asarray(frames_list)
    return frames_list


def read_avi(filename):
    """
    Reading video with AVI format.

    Parameters
    ----------
    file_name: str
        Path and name of AVI video.

    Returns
    -------
    video: NDArray
        The video is 3D-numpy (number of frames, width, height).
    """

    cap = cv2.VideoCapture(filename)
    frames_list = []
    while(cap.isOpened()):
        ret, frame = cap.read()
        if ret and frame is not None:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            frames_list.append(gray)
        else:
            break

    cap.release()
    video = np.asarray(frames_list)
    return video


def read_png(filename):
    """
    Reading image with PNG format.

    Parameters
    ----------
    file_name: str
        Path and name of PNG image.

    Returns
    -------
    @returns: NDArray
        The video is 2D-numpy (width, height).
    """
    img = io.imread(filename)
    if len(img.shape) == 3:
        grayscale = rgb2gray(img)
    elif len(img.shape) == 2:
        grayscale = img

    return grayscale


def read_fits(filename):
    """
    Reading image/video with fits format.

    Parameters
    ----------
    file_name: str
        Path and name of fits image/video.

    Returns
    -------
    @returns: NDArray
        The video as 3D-numpy with the following shape (number of frames, width, height)

    """
    hdul = fits.open(filename)
    return hdul[0].data


def read_fli(filename):
    """
    Reading image/video with fli format.

    Parameters
    ----------
    file_name: str
        Path and name of fits image/video.

    Returns
    -------
    @returns: NDArray
        The video as 3D-numpy with the following shape (number of frames, width, height)

    """
    myflifile = FliFile(filename)
    data = myflifile.getdata()
    data_ = np.transpose(data, (2, 0, 1))
    return data_



class DirectoryType:

    def __init__(self, dirName, type_file):
        """
        Based on the type file description, this class generates a dataframe
        containing 'Directory,' 'Folder,' and 'File' from all files below the define directory.

        Parameters
        ----------
        dirName: str
            A directory that is used to look for files of a particular type file.

        type_file: str
            The type of files that the user is looking for.
        """
        self.dirName = dirName
        self.type = type_file

        # Get the list of all files in directory tree at given path
        fileNames = []
        dirPaths = []
        dirnames = []
        for (dirpath, dirname, filenames) in os.walk(dirName):
            for file in filenames:
                if file.endswith(self.type):
                    dirPaths.append(dirpath)
                    fileNames.append(file)
                    dirnames.append(os.path.basename(dirpath))

        self.df = pd.DataFrame(list(zip(dirPaths, dirnames, fileNames)), columns=['Directory', 'Folder', 'File'])

    def return_df(self):
        """
        Based on the type file specification, this function returns a pandas data frame containing
        'Directory,"Folder,' and 'File' from all files below the define directory.

        Returns
        -------
            The data frame contains ('Directory', 'Folder', 'File')
        """
        return self.df

    def get_list_of_files(self, dirName):
        # create a list of file and sub directories
        # names in the given directory
        listOfFile = os.listdir(dirName)
        allFiles = list()
        # Iterate over all the entries
        for entry in listOfFile:
            # Create full path
            fullPath = os.path.join(dirName, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(fullPath):
                allFiles = allFiles + self.get_list_of_files(fullPath)
            else:
                allFiles.append(fullPath)

        return allFiles








from piscat.BackgroundCorrection.noise_floor import NoiseFloor
from piscat.Preproccessing import Normalization

from PySide6 import QtGui, QtCore, QtWidgets
import pandas as pd
import os
import matplotlib.pyplot as plt
import matplotlib.patches as patches


class Noise_Floor(QtWidgets.QMainWindow):

    def __init__(self, video, parent=None):
        super(Noise_Floor, self).__init__(parent)
        self.window = QtWidgets.QWidget()

        self.original_video = video
        self.original_video_pn = None

        self.min_radius = None
        self.max_radius = None
        self.step_radius = None
        self.radius_list = None
        self.mode = None
        self.file_path = None
        self.find_radius_update_tab_flag = True

        # self.checkbox_power_normalization = QtWidgets.QCheckBox("Laser power normalization", self)

        self.groupBox_PN = QtWidgets.QGroupBox("Laser Power normalization:")
        self.groupBox_PN.setCheckable(True)
        self.groupBox_PN.setChecked(False)

        self.start_win_size_x = QtWidgets.QLineEdit()
        self.start_win_size_x.setPlaceholderText('ROI X start')
        self.start_win_size_x_label = QtWidgets.QLabel("start width pixel:")

        self.end_win_size_x = QtWidgets.QLineEdit()
        self.end_win_size_x.setPlaceholderText('ROI X end')
        self.end_win_size_x_label = QtWidgets.QLabel("end width pixel:")

        self.start_win_size_y = QtWidgets.QLineEdit()
        self.start_win_size_y.setPlaceholderText('ROI Y start')
        self.start_win_size_y_label = QtWidgets.QLabel("start hight pixel:")

        self.end_win_size_y = QtWidgets.QLineEdit()
        self.end_win_size_y.setPlaceholderText('ROI X end')
        self.end_win_size_y_label = QtWidgets.QLabel("start hight pixel:")

        self.selected_frame = QtWidgets.QLineEdit()
        self.selected_frame.setPlaceholderText('selected frame')
        self.selected_frame_label = QtWidgets.QLabel("selected preview frame:")
        self.selected_frame.setText('0')

        self.preview_roi = QtWidgets.QPushButton("ROI preview")
        self.preview_roi.setAutoDefault(False)
        self.preview_roi.clicked.connect(self.preview_roi_plot)
        self.preview_roi.setFixedHeight(20)
        self.preview_roi.setFixedWidth(100)

        grid_pn = QtWidgets.QGridLayout()
        grid_pn.addWidget(self.start_win_size_x_label, 0, 0)
        grid_pn.addWidget(self.start_win_size_x, 0, 1)

        grid_pn.addWidget(self.end_win_size_x_label, 0, 2)
        grid_pn.addWidget(self.end_win_size_x, 0, 3)

        grid_pn.addWidget(self.start_win_size_y_label, 1, 0)
        grid_pn.addWidget(self.start_win_size_y, 1, 1)

        grid_pn.addWidget(self.end_win_size_y_label, 1, 2)
        grid_pn.addWidget(self.end_win_size_y, 1, 3)

        grid_pn.addWidget(self.selected_frame_label, 2, 0)
        grid_pn.addWidget(self.selected_frame, 2, 1)
        grid_pn.addWidget(self.preview_roi, 2, 2)

        self.groupBox_PN.setLayout(grid_pn)

        self.checkbox_save = QtWidgets.QCheckBox("Saving as CSV", self)
        self.checkbox_save.toggled.connect(lambda: self.save_active())

        self.checkbox_loglog_scale = QtWidgets.QRadioButton("log scale")
        self.checkbox_loglog_scale.setChecked(True)
        self.checkbox_normal_scale = QtWidgets.QRadioButton("normal scale")

        self.ok = QtWidgets.QPushButton("Ok")
        self.ok.setAutoDefault(False)
        self.ok.clicked.connect(self.do_update)
        self.ok.setFixedWidth(100)

        self.plot = QtWidgets.QPushButton("Plot")
        self.plot.setAutoDefault(False)
        self.plot.clicked.connect(self.do_plot)
        self.plot.setFixedWidth(100)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)
        self.grid.addWidget(self.createSecondExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.createThirdExclusiveGroup(), 2, 0)
        self.grid.addWidget(self.groupBox_PN, 5, 0)
        self.grid.addWidget(self.checkbox_save, 6, 0)
        self.grid.addWidget(self.checkbox_loglog_scale, 7, 0)
        self.grid.addWidget(self.checkbox_normal_scale, 8, 0)
        self.grid.addWidget(self.ok, 9, 0)
        self.grid.addWidget(self.plot, 10, 0)

        self.setWindowTitle("Noise Floor")
        self.window.setLayout(self.grid)
        self.window.show()

    def __del__(self):
        del self
        print('Destructor called, Employee deleted.')

    def createFirstExclusiveGroup(self):
        self.groupBox_range = QtWidgets.QGroupBox("Give range for batch sizes:")
        self.groupBox_range.setCheckable(True)
        self.groupBox_range.setChecked(False)
        self.grid1 = QtWidgets.QGridLayout()
        self.add_line_edit1()
        self.groupBox_range.setLayout(self.grid1)

        return self.groupBox_range

    def createSecondExclusiveGroup(self):
        self.groupBox_list = QtWidgets.QGroupBox("Give list for batch sizes")
        self.groupBox_list.setCheckable(True)
        self.groupBox_list.setChecked(False)
        self.grid2 = QtWidgets.QGridLayout()
        self.add_line_edit2()
        self.groupBox_list .setLayout(self.grid2)

        return self.groupBox_list

    def createThirdExclusiveGroup(self):
        self.groupBox_FPNc = QtWidgets.QGroupBox("FPNc:")
        self.groupBox_FPNc.setCheckable(True)
        self.groupBox_FPNc.setChecked(False)

        self.FPN_mode_group = QtWidgets.QButtonGroup()
        self.radio_wFPN_mode = QtWidgets.QRadioButton("Wavelet FPNc")
        self.radio_mFPN_mode = QtWidgets.QRadioButton("Median FPNc")
        self.radio_cpFPN_mode = QtWidgets.QRadioButton("Column projection FPNc")
        self.radio_fft_FPN_mode = QtWidgets.QRadioButton("FFT2D FPNc")

        self.FPN_mode_group.addButton(self.radio_wFPN_mode)
        self.FPN_mode_group.addButton(self.radio_mFPN_mode)
        self.FPN_mode_group.addButton(self.radio_cpFPN_mode)
        self.FPN_mode_group.addButton(self.radio_fft_FPN_mode)

        self.radio_wFPN_mode.toggled.connect(self.axis_selection)
        self.radio_mFPN_mode.toggled.connect(self.axis_selection)
        self.radio_cpFPN_mode.toggled.connect(self.axis_selection)
        self.radio_fft_FPN_mode.toggled.connect(self.axis_selection)

        self.axis_group = QtWidgets.QButtonGroup()
        self.radio_axis_1 = QtWidgets.QRadioButton("FPNc in axis 0")
        self.radio_axis_2 = QtWidgets.QRadioButton("FPNc in axis 1")
        self.radio_axis_3 = QtWidgets.QRadioButton("FPNc in Both axis")
        self.radio_axis_2.setChecked(True)
        self.radio_mFPN_mode.setChecked(True)

        self.axis_group.addButton(self.radio_axis_1)
        self.axis_group.addButton(self.radio_axis_2)
        self.axis_group.addButton(self.radio_axis_3)

        grid = QtWidgets.QGridLayout()
        grid.addWidget(self.radio_mFPN_mode, 1, 0)
        grid.addWidget(self.radio_cpFPN_mode, 2, 0)
        grid.addWidget(self.radio_wFPN_mode, 3, 0)
        grid.addWidget(self.radio_fft_FPN_mode, 4, 0)

        grid.addWidget(self.radio_axis_1, 1, 1)
        grid.addWidget(self.radio_axis_2, 2, 1)
        grid.addWidget(self.radio_axis_3, 3, 1)

        self.groupBox_FPNc.setLayout(grid)
        return self.groupBox_FPNc

    def axis_selection(self):
        if self.radio_mFPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_wFPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_fft_FPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_cpFPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

    @QtCore.Slot()
    def do_update(self):

        if self.ok.clicked:
            self.get_values()
            if self.groupBox_range.isChecked() and not(self.groupBox_list.isChecked()):

                if self.min_radius != '' and self.max_radius != '' and self.step_radius != '':
                    self.min_radius = int(self.min_radius)
                    self.max_radius = int(self.max_radius)
                    self.step_radius = int(self.step_radius)
                    self.radius_list = list(range(self.min_radius, self.max_radius, self.step_radius))
                    self.mode = 'Range'
                    self.find_radius_update_tab_flag = False
                    self.run_noiseFloor()

                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Please insert radius size!")
                    self.msg_box.exec_()

            if self.groupBox_list.isChecked() and not(self.groupBox_range.isChecked()):

                if self.radius_list != '':
                    self.radius_list = eval(self.radius_list)
                    self.mode = 'List'
                    self.find_radius_update_tab_flag = False
                    self.run_noiseFloor()
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Please insert radius size!")
                    self.msg_box.exec_()

            elif self.groupBox_list.isChecked() and self.groupBox_range.isChecked():
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Only one method should active!")
                self.msg_box.exec_()

            elif not(self.groupBox_list.isChecked()) and not(self.groupBox_range.isChecked()):
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Please select one of the methods!")
                self.msg_box.exec_()

    def run_noiseFloor(self):

        if self.groupBox_PN.isChecked():
            try:
                self.s_x_win_pn = int(self.start_win_size_x.text())
                self.e_x_win_pn = int(self.end_win_size_x.text())
                self.s_y_win_pn = int(self.start_win_size_y.text())
                self.e_y_win_pn = int(self.end_win_size_y.text())
            except:
                self.s_x_win_pn = None
                self.e_x_win_pn = None
                self.s_y_win_pn = None
                self.e_y_win_pn = None

            self.original_video_pn, _ = Normalization(video=self.original_video).power_normalized(roi_x=(self.s_x_win_pn, self.e_x_win_pn),
                                                                                                roi_y=(self.s_y_win_pn, self.e_y_win_pn),)
        else:
            self.original_video_pn = self.original_video

        if self.groupBox_FPNc.isChecked():
            FPN_flag = True
            if self.radio_axis_1.isChecked():
                self.axis = 0
            elif self.radio_axis_2.isChecked():
                self.axis = 1
            elif self.radio_axis_3.isChecked():
                self.axis = 'Both'

            if self.radio_mFPN_mode.isChecked():
                self.mode_FPN = 'mFPN'
            elif self.radio_cpFPN_mode.isChecked():
                self.mode_FPN = 'cpFPN'
            elif self.radio_wFPN_mode.isChecked():
                self.mode_FPN = 'wFPN'
            elif self.radio_fft_FPN_mode.isChecked():
                self.mode_FPN = 'fFPN'

        else:
            FPN_flag = False
            self.axis = None
            self.mode_FPN = 'mFPN'

        result_flag = True
        n_jobs = os.cpu_count()
        inter_flag_parallel_active = True
        flag_first_except = False
        while result_flag:
            try:
                self.noise_floor_ = NoiseFloor(self.original_video_pn, list_range=self.radius_list,
                                               select_correction_axis=self.axis,
                                               FPN_flag=FPN_flag, mode_FPN=self.mode_FPN, n_jobs=None,
                                               inter_flag_parallel_active=inter_flag_parallel_active)

                if self.checkbox_loglog_scale.isChecked():
                    self.noise_floor_.plot_result(flag_log=True)
                else:
                    self.noise_floor_.plot_result(flag_log=False)

                if self.checkbox_save.isChecked() and self.file_path is not None:

                    noise_floor = {'batch size': self.noise_floor_.list_range, 'SNR': self.noise_floor_.mean}
                    noise_floor_df = pd.DataFrame(data=noise_floor)
                    noise_floor_df.to_csv(self.file_path)
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Nothing Save!")
                    self.msg_box.exec_()

                result_flag = False
            except:
                inter_flag_parallel_active = False
                if flag_first_except:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Memory Error! Please off parallel CPU.")
                    self.msg_box.exec_()
                    result_flag = False
                flag_first_except = True

    def do_plot(self):
        try:
            if self.checkbox_loglog_scale.isChecked():
                self.noise_floor_.plot_result(flag_log=True)
            else:
                self.noise_floor_.plot_result(flag_log=False)
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please update first!5")
            self.msg_box.exec_()

    def preview_roi_plot(self, selected_frame):
        if self.groupBox_PN.isChecked():
            self.flag_power_normalization = True
            try:
                self.s_x_win_pn = int(self.start_win_size_x.text())
                self.e_x_win_pn = int(self.end_win_size_x.text())
                self.s_y_win_pn = int(self.start_win_size_y.text())
                self.e_y_win_pn = int(self.end_win_size_y.text())
                selected_frame = int(self.selected_frame.text())

                img_ = self.original_video[selected_frame, :, :]
                roi_img_ = img_[self.s_x_win_pn:self.e_x_win_pn, self.s_y_win_pn:self.e_y_win_pn]

                fig, ax = plt.subplots()
                ax.imshow(img_, cmap='gray')
                rect = patches.Rectangle((self.s_x_win_pn, self.s_y_win_pn), roi_img_.shape[0], roi_img_.shape[1]
                                         , linewidth=2, edgecolor='r', facecolor='none')
                ax.add_patch(rect)
                plt.show()

            except:
                self.msg_box1 = QtWidgets.QMessageBox()
                self.msg_box1.setWindowTitle("Warning!")
                self.msg_box1.setText("The ROI is not defined!")
                self.msg_box1.exec_()

    def get_values(self):

        if self.groupBox_range.isChecked():
            self.min_radius = self.le1.text()
            self.max_radius = self.le2.text()
            self.step_radius = self.le3.text()

        elif self.groupBox_list.isChecked():
            self.radius_list = self.le4.text()

    def add_line_edit1(self):
        self.le1 = QtWidgets.QLineEdit()
        self.le1.setPlaceholderText('Min. batch size')
        self.le_1_label = QtWidgets.QLabel("Min batch size:")

        self.le2 = QtWidgets.QLineEdit()
        self.le2.setPlaceholderText('Max. batch size')
        self.le_2_label = QtWidgets.QLabel("Max batch size:")

        self.le3 = QtWidgets.QLineEdit()
        self.le3.setPlaceholderText('step')
        self.le_3_label = QtWidgets.QLabel("Stride between batch size:")

        self.grid1.addWidget(self.le_1_label, 2, 0)
        self.grid1.addWidget(self.le_2_label, 3, 0)
        self.grid1.addWidget(self.le_3_label, 4, 0)

        self.grid1.addWidget(self.le1, 2, 1)
        self.grid1.addWidget(self.le2, 3, 1)
        self.grid1.addWidget(self.le3, 4, 1)

    def add_line_edit2(self):
        self.le4 = QtWidgets.QLineEdit()
        self.le4.setPlaceholderText('list of batches size')
        self.le_4_label = QtWidgets.QLabel("List of all batches")

        self.grid2.addWidget(self.le_4_label, 0, 0)
        self.grid2.addWidget(self.le4, 1, 1)

    def save_active(self):
        if self.checkbox_save.isChecked():
            self.file_path = False

            self.file_path, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Noise floor",
                                                                      QtCore.QDir.currentPath())

            self.file_path = self.file_path + '_noise_floor.csv'

    def closeEvent(self, event):
        event.accept()  # let the window close





from .filtering import FFT2D
from .filtering import Filters
from .filtering import RadialVarianceTransform
from .filtering import FastRadialSymmetryTransform
from .filtering import GuidedFilter
from .filtering import GrayGuidedFilter
from .filtering import MultiDimGuidedFilter
from .normalization import Normalization
from .patch_genrator import ImagePatching
from .FPNc import ColumnProjectionFPNc
from .FPNc import MedianProjectionFPNc
from .FPNc import FrequencyFPNc




from PySide6 import QtGui, QtCore, QtWidgets


class Cropping(QtWidgets.QMainWindow):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.width_size_s = None
        self.width_size_e = None

        self.height_size_s = None
        self.height_size_e = None

        self.height_size = -1
        self.width_size = -1

        self.frame_s = None
        self.frame_e = None
        self.frame_jump = None

        self.type = None

        self.flag_display = False
        self.flag_RGB2GRAY = False
        self.raw_data_update_flag = True

        self.window = QtWidgets.QWidget()

        self.checkbox_display = QtWidgets.QCheckBox("Display", self)
        self.checkbox_RGB2GRAY = QtWidgets.QCheckBox("RGB2Gray", self)

        self.ok = QtWidgets.QPushButton("Ok")
        self.ok.setAutoDefault(False)
        self.ok.setFixedWidth(100)
        self.ok.clicked.connect(self.do_update)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createSecondExclusiveGroup(), 0, 0)

        self.grid.addWidget(self.checkbox_display, 5, 0)
        self.grid.addWidget(self.checkbox_RGB2GRAY, 6, 0)
        self.grid.addWidget(self.ok, 7, 0)

        self.setWindowTitle("Reading video")
        self.setStyleSheet('QMainWindow{background-color: darkgray;}')
        self.window.setLayout(self.grid)
        self.window.show()

    def closeEvent(self, **kwargs):
        QtCore.QCoreApplication.instance().quit()

    def createSecondExclusiveGroup(self):

        self.groupBox_cropping = QtWidgets.QGroupBox("Video Cropping")
        self.groupBox_cropping.setCheckable(True)
        self.groupBox_cropping.setChecked(False)
        self.grid2 = QtWidgets.QGridLayout()
        self.info_cut_edith()
        self.groupBox_cropping.setLayout(self.grid2)

        return self.groupBox_cropping

    def do_update(self):
        if self.ok.clicked:
            self.get_values()
            if self.checkbox_display.isChecked():
                self.flag_display = True

            if self.checkbox_RGB2GRAY.isChecked():
                self.flag_RGB2GRAY = True

            if self.groupBox_cropping.isChecked():
                if self.width_size_s != '' and self.width_size_e != '':
                    self.width_size_s = int(self.width_size_s)
                    self.width_size_e = int(self.width_size_e)
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("All rows are selected")
                    self.msg_box.exec_()
                    self.width_size_s = 0
                    self.width_size_e = self.width_size

                if self.height_size_s != '' and self.height_size_e != '':
                    self.height_size_s = int(self.height_size_s)
                    self.height_size_e = int(self.height_size_e)
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("All columns are selected")
                    self.msg_box.exec_()
                    self.height_size_s = 0
                    self.height_size_e = self.height_size

                if self.frame_s != '' and self.frame_e != '' :
                    if self.frame_jump != '':
                        self.frame_s = int(self.frame_s)
                        self.frame_e = int(self.frame_e)
                        self.frame_jump = int(self.frame_jump)
                    else:
                        self.frame_s = int(self.frame_s)
                        self.frame_e = int(self.frame_e)
                        self.frame_jump = 1
                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("All frames are selected")
                    self.msg_box.exec_()
                    self.frame_s = 0
                    self.frame_e = -1
                    self.frame_jump = 1


            self.raw_data_update_flag = False
            self.window.close()

    def check_cut_size_odd(self):

        if (self.width_size_s % 2) != 0:
            self.width_size_s = self.width_size_s + 1

        if (self.width_size_e % 2) != 0:
            self.width_size_e = self.width_size_e - 1

        if (self.height_size_s % 2) != 0:
            self.height_size_s = self.height_size_s + 1

        if (self.height_size_e % 2) != 0:
            self.height_size_e = self.height_size_e - 1

    def info_cut_edith(self):
        self.le3 = QtWidgets.QLineEdit()
        self.le3.setPlaceholderText('width_start')
        self.le_3_label = QtWidgets.QLabel("start width pixel:")

        self.le4 = QtWidgets.QLineEdit()
        self.le4.setPlaceholderText('width_end')
        self.le_4_label = QtWidgets.QLabel("end width pixel:")

        self.le5 = QtWidgets.QLineEdit()
        self.le5.setPlaceholderText('height_start')
        self.le_5_label = QtWidgets.QLabel("start height pixel:")

        self.le6 = QtWidgets.QLineEdit()
        self.le6.setPlaceholderText('height_end')
        self.le_6_label = QtWidgets.QLabel("end height pixel:")

        self.le7 = QtWidgets.QLineEdit()
        self.le7.setPlaceholderText('frame_start')
        self.le_7_label = QtWidgets.QLabel("start frame:")

        self.le8 = QtWidgets.QLineEdit()
        self.le8.setPlaceholderText('frame_end')
        self.le_8_label = QtWidgets.QLabel("end frame:")

        self.le9 = QtWidgets.QLineEdit()
        self.le9.setPlaceholderText('frame_jump')
        self.le_9_label = QtWidgets.QLabel("frame stride:")

        self.grid2.addWidget(self.le_3_label, 1, 0)
        self.grid2.addWidget(self.le_4_label, 1, 2)
        self.grid2.addWidget(self.le_5_label, 2, 0)
        self.grid2.addWidget(self.le_6_label, 2, 2)
        self.grid2.addWidget(self.le_7_label, 3, 0)
        self.grid2.addWidget(self.le_8_label, 3, 2)
        self.grid2.addWidget(self.le_9_label, 3, 4)

        self.grid2.addWidget(self.le3, 1, 1)
        self.grid2.addWidget(self.le4, 1, 3)
        self.grid2.addWidget(self.le5, 2, 1)
        self.grid2.addWidget(self.le6, 2, 3)
        self.grid2.addWidget(self.le7, 3, 1)
        self.grid2.addWidget(self.le8, 3, 3)
        self.grid2.addWidget(self.le9, 3, 5)

    def get_values(self):
        if self.groupBox_cropping.isChecked():
            self.width_size_s = self.le3.text()
            self.width_size_e = self.le4.text()

            self.height_size_s = self.le5.text()
            self.height_size_e = self.le6.text()

            self.frame_s = self.le7.text()
            self.frame_e = self.le8.text()
            self.frame_jump = self.le9.text()

    def handleClose(self):
        print("closing Tab B")



"""
__author__ = "Houman Mirzaalian D., xxxx"
"""

from __future__ import print_function

import numpy as np
import matplotlib.ticker as ticker
import pylab as pl
from PySide6 import QtGui
from PySide6 import QtCore
from PySide6 import QtWidgets
from PySide6.QtCore import *
from scipy.ndimage import median_filter
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from matplotlib import animation
from matplotlib.patches import Circle, Arrow, Rectangle
from piscat.InputOutput import read_status_line
from piscat.Trajectory.data_handeling import protein_trajectories_list2dic
from piscat.Visualization.print_colors import PrintColors


class WorkerSignals(QObject):

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class HTML_Display():

    def __init__(self, video, color='gray', time_delay=0.5, median_filter_flag=False, imgSizex=5, imgSizey=5):
        """
        This class display the video for HTML.

        Parameters
        ----------
        video:(NDArray)
           Input video

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        time_delay: float
           Delay between frames in milliseconds.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.
        """
        self.video = video
        self.median_filter_flag = median_filter_flag

        self.fig = plt.figure()
        self.fig.set_size_inches(imgSizex, imgSizey, True)
        self.ax = self.fig.add_subplot(111)
        self.pressed_key = {}
        plt.axis('off')

        self.div = make_axes_locatable(self.ax)
        self.cax = self.div.append_axes('right', '5%', '5%')

        self.cv0 = self.video[0, :, :]
        self.im = self.ax.imshow(self.cv0, origin='lower', cmap=color)
        self.cb = self.fig.colorbar(self.im, cax=self.cax)

        self.tx = self.ax.set_title('Frame 0')

        self.ani = animation.FuncAnimation(self.fig, self.animate, frames=self.video.shape[0],
                                           blit=False, interval=time_delay, repeat=False, cache_frame_data=False)

    def animate(self, i_):

        arr = self.video[i_, :, :]
        vmax = np.max(arr)
        vmin = np.min(arr)
        if self.median_filter_flag:
            frame_v = median_filter(arr, 3)
        else:
            frame_v = arr
        self.im.set_data(frame_v)
        self.im.set_clim(vmin, vmax)
        self.tx.set_text('Frame {0}'.format(i_))


class HTML_Display_StatusLine():

    def __init__(self, video, median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, time_delay=0.5):
        """
        This class displays the video in the HTML while highlight status line

        Parameters
        ----------
        video: NDArray
           Input video.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        time_delay: float
           Delay between frames in milliseconds.
       """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.median_filter_flag = median_filter_flag

        status_ = read_status_line.StatusLine(video)
        _, status_information = status_.find_status_line()
        self.statusLine_position = status_information['status_line_position']

        self.fig = plt.figure()
        self.fig.set_size_inches(imgSizex, imgSizey, True)
        self.ax = self.fig.add_subplot(111)
        self.pressed_key = {}
        plt.axis('off')

        self.div = make_axes_locatable(self.ax)
        self.cax = self.div.append_axes('right', '5%', '5%')

        self.cv0 = self.video[0, :, :]
        self.im = self.ax.imshow(self.cv0, origin='lower', cmap=self.color)
        self.cb = self.fig.colorbar(self.im, cax=self.cax)
        self.tx = self.ax.set_title('Frame 0')

        if self.statusLine_position == 'column':
            rect = Rectangle((0, self.video.shape[2]), self.video.shape[2], 1, linewidth=10, edgecolor='r',
                                  facecolor='none')
            self.ax.add_patch(rect)
        elif self.statusLine_position == 'row':
            rect = Rectangle((self.video.shape[1], 0), 1, self.video.shape[1], linewidth=10, edgecolor='r',
                                  facecolor='none')
            self.ax.add_patch(rect)

        self.ani = animation.FuncAnimation(self.fig, self.animate, frames=self.video.shape[0],
                                           blit=False, interval=time_delay, repeat=False, cache_frame_data=False)

    def animate(self, i_):
        [p.remove() for p in reversed(self.ax.patches)]
        arr = self.video[i_, :, :]
        vmax = np.max(arr)
        vmin = np.min(arr)
        if self.median_filter_flag:
            frame_v = median_filter(arr, 3)
        else:
            frame_v = arr
        self.im.set_data(frame_v)
        self.im.set_clim(vmin, vmax)
        self.tx.set_text('Frame {0}'.format(i_))

        if self.statusLine_position == 'column':
            rect = Rectangle((0, self.video.shape[2]), self.video.shape[2], 1, linewidth=10, edgecolor='r',
                                  facecolor='none')
            self.ax.add_patch(rect)

        elif self.statusLine_position == 'row':
            rect = Rectangle((self.video.shape[1], 0), 1, self.video.shape[1], linewidth=10, edgecolor='r',
                                  facecolor='none')
            self.ax.add_patch(rect)


class HTML_PSFs_subplotLocalizationDisplay(QRunnable, PrintColors):

    def __init__(self, list_videos, list_df_PSFs, list_titles, numRows, numColumns, color='gray',
                 median_filter_flag=False, imgSizex=5, imgSizey=5, time_delay=0.5, save_path=None, fps=10):
        """
        This class displays the videos in the HTML while highlight PSFs.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos.

        list_df_PSFs: list panda data_frame
            List Data Frames that contains the location of PSFs for each video.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        time_delay: float
                   Delay between frames in milliseconds.
        """
        super(HTML_PSFs_subplotLocalizationDisplay, self).__init__()
        PrintColors.__init__(self)

        self.list_video = list_videos
        self.list_df_PSFs = list_df_PSFs
        self.save_path = save_path
        self.fps = fps
        self.signals = WorkerSignals()
        self.median_filter_flag = median_filter_flag
        self.memory = 0
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.imgGrid_list = []
        for i_ in range(1, self.numRows*self.numColumns+1):
            img_grid_ = self.fig.add_subplot(self.numRows, self.numColumns, i_)
            img_grid_.axis('off')
            self.imgGrid_list.append(img_grid_)

        self.fig.tight_layout()
        self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        for ax_, vid_, df_PSFs, tit_ in zip(self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles):
            div_ = make_axes_locatable(ax_)
            self.div_0.append(div_)

            cax_ = div_.append_axes('right', '5%', '5%')
            self.cax_0.append(cax_)

            im_ = ax_.imshow(vid_[0, :, :], origin='lower', cmap=color)
            self.im_0.append(im_)
            self.cb_0.append(self.fig.colorbar(im_, cax=cax_))

            if tit_ is not None:
                self.tx_0.append(ax_.set_title(tit_ + ', Frame 0'))
            else:
                self.tx_0.append(ax_.set_title('Frame 0'))

            particle = df_PSFs.loc[df_PSFs['frame'] == 0]
            particle_X = particle['x'].tolist()
            particle_Y = particle['y'].tolist()
            particle_sigma = particle['sigma'].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax_.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))

        self.ani = animation.FuncAnimation(self.fig, self.animate, frames=self.list_video[0].shape[0],
                                           blit=False, interval=time_delay, repeat=False, cache_frame_data=False)

    def animate(self, i_):
        if i_ < self.list_video[0].shape[0]:
            for idx_, (ax_, vid_, df_PSFs, tit_) in enumerate(zip(self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles)):
                [p.remove() for p in reversed(ax_.patches)]

                arr = vid_[i_, :, :]
                vmax = np.max(arr)
                vmin = np.min(arr)
                if self.median_filter_flag:
                    frame_v = median_filter(arr, 3)
                else:
                    frame_v = arr

                self.im_0[idx_].set_data(frame_v)
                self.im_0[idx_].set_clim(vmin, vmax)

                if tit_ is not None:
                    self.tx_0[idx_].set_text(tit_ + ', Frame {}'.format(i_))
                else:
                    self.tx_0[idx_].set_text('Frame {}'.format(i_))

                particle = df_PSFs.loc[df_PSFs['frame'] == i_]
                particle_X = particle['x'].tolist()
                particle_Y = particle['y'].tolist()
                particle_sigma = particle['sigma'].tolist()
                for j_ in range(len(particle_X)):

                    y = int(particle_Y[j_])
                    x = int(particle_X[j_])
                    sigma = particle_sigma[j_]
                    ax_.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))

    @Slot()
    def run(self, *args, **kwargs):
        if self.save_path is not None:
            dpi = 300
            writer = animation.writers['ffmpeg'](fps=self.fps, bitrate=-1, extra_args=['-vcodec', 'libx264'])
            self.ani.save(self.save_path, writer=writer, dpi=dpi)

            self.signals.finished.emit()
        else:
            print(f"{self.WARNING}\nError occurred in MP4!{self.ENDC}")
            self.signals.finished.emit()


class HTML_subplotDisplay():

    def __init__(self, list_videos, list_titles, numRows, numColumns, color='gray',
                 median_filter_flag=False, imgSizex=5, imgSizey=5, time_delay=0.5):
        """
        This class displays the videos in the HTML.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        time_delay: float
            Delay between frames in milliseconds.
        """

        self.list_video = list_videos
        self.median_filter_flag = median_filter_flag
        self.memory = 0
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.imgGrid_list = []
        for i_ in range(1, self.numRows*self.numColumns+1):
            img_grid_ = self.fig.add_subplot(self.numRows, self.numColumns, i_)
            img_grid_.axis('off')
            self.imgGrid_list.append(img_grid_)

        self.fig.tight_layout()
        self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        for ax_, vid_, tit_ in zip(self.imgGrid_list, self.list_video, self.list_titles):
            div_ = make_axes_locatable(ax_)
            self.div_0.append(div_)

            cax_ = div_.append_axes('right', '5%', '5%')
            self.cax_0.append(cax_)

            im_ = ax_.imshow(vid_[0, :, :], origin='lower', cmap=color)
            self.im_0.append(im_)
            self.cb_0.append(self.fig.colorbar(im_, cax=cax_))

            if tit_ is not None:
                self.tx_0.append(ax_.set_title(tit_ + ', Frame 0'))
            else:
                self.tx_0.append(ax_.set_title('Frame 0'))

        self.ani = animation.FuncAnimation(self.fig, self.animate, frames=self.list_video[0].shape[0],
                                           blit=False, interval=time_delay, repeat=False, cache_frame_data=False)

    def animate(self, i_):
        if i_ < self.list_video[0].shape[0]:
            for idx_, (ax_, vid_, tit_) in enumerate(zip(self.imgGrid_list, self.list_video, self.list_titles)):

                arr = vid_[i_, :, :]
                vmax = np.max(arr)
                vmin = np.min(arr)
                if self.median_filter_flag:
                    frame_v = median_filter(arr, 3)
                else:
                    frame_v = arr

                self.im_0[idx_].set_data(frame_v)
                self.im_0[idx_].set_clim(vmin, vmax)

                if tit_ is not None:
                    self.tx_0[idx_].set_text(tit_ + ', Frame {}'.format(i_))
                else:
                    self.tx_0[idx_].set_text('Frame {}'.format(i_))


class HTMLSelectedPSFs_localizationDisplay():

    def __init__(self, video, particles, particles_num='#0', frame_extend=0, color='gray',
                 median_filter_flag=False, imgSizex=5, imgSizey=5, time_delay=0.5):
        """
        This class displays video in the HTML while highlighting selected PSF.

        Parameters
        ----------
        video: NDArray
           Input video.

        particles: dic
             Dictionary similar to the following structures.:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ..., 'particle_ID': ...},
                "#1": {}, ...}

        particles_num: str
            Choose the corresponding key in the particles dictionary.

        frame_extend: int
            Display particle for ``frame_extend`` before and after segmented ones. In case there are not enough frames before/after
            , it shows only for the number of existing frames.

        flag_fit2D: bool
            It activate 2D-Gaussian fit to extract fitting information of selected PSF.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        time_delay: float
            Delay between frames in milliseconds.
        """

        self.video = video
        self.particles = particles
        self.median_filter_flag = median_filter_flag
        self.memory = 0

        if type(particles) is list:
            self.particles = protein_trajectories_list2dic(particles)

        if type(self.particles) is dict:
            for key, item in self.particles.items():
                if key == particles_num:
                    intensity_horizontal = item['intensity_horizontal']
                    intensity_vertical = item['intensity_vertical']
                    center_int = item['center_int']
                    center_int_flow = item['center_int_flow']
                    self.frame_number_ = item['frame_number']
                    self.sigma = item['sigma']
                    self.x_center = item['x_center']
                    self.y_center = item['y_center']
                    self.particle_ID = item['particle_ID']

        video_frameNum = video.shape[0]
        max_particle_frame = np.max(self.frame_number_)
        min_particle_frame = np.min(self.frame_number_)

        max_extend_particle_frame = np.min([max_particle_frame + frame_extend, video_frameNum])
        min_extend_particle_frame = np.max([min_particle_frame - frame_extend, 0])

        self.frame_number = list(range(min_extend_particle_frame, max_extend_particle_frame, 1))

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
        self.imgGrid_list = []

        img_grid_ = self.fig.add_subplot(1, 1, 1)
        img_grid_.axis('off')
        self.imgGrid_list = img_grid_

        self.fig.tight_layout()
        # self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        div_ = make_axes_locatable(self.imgGrid_list)
        self.div_0.append(div_)

        cax_ = div_.append_axes('right', '5%', '5%')
        self.cax_0.append(cax_)

        selected_frame = self.frame_number[0]
        im_ = self.imgGrid_list.imshow(video[selected_frame, :, :], origin='lower', cmap=color)
        self.im_0 = im_
        self.cb_0 = self.fig.colorbar(im_, cax=cax_)

        self.tx_0 = self.imgGrid_list.set_title('Frame 0')

        if 0 <= np.max(self.frame_number_) and 0 >= np.min(self.frame_number_):
            idx_ = np.where(self.frame_number_ == 0)

            particle_X = self.x_center[idx_]
            particle_Y = self.y_center[idx_]
            particle_sigma = self.sigma[idx_]
            particle_labels = self.particle_ID[idx_]

            y = int(particle_Y)
            x = int(particle_X)
            sigma = particle_sigma
            self.imgGrid_list.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))

        self.ani = animation.FuncAnimation(self.fig, self.animate, frames=self.frame_number,
                                           blit=False, interval=time_delay, repeat=False, cache_frame_data=False)

    def animate(self, i_):
        if i_ <= np.max(self.frame_number) and i_ >= np.min(self.frame_number):
            [p.remove() for p in reversed(self.imgGrid_list.patches)]

            arr = self.video[i_, :, :]
            vmax = np.max(arr)
            vmin = np.min(arr)
            if self.median_filter_flag:
                frame_v = median_filter(arr, 3)
            else:
                frame_v = arr

            self.im_0.set_data(frame_v)
            self.im_0.set_clim(vmin, vmax)
            self.tx_0.set_text('Frame {}'.format(i_))

            if i_ <= np.max(self.frame_number_) and i_ >= np.min(self.frame_number_):
                idx_ = np.where(self.frame_number_ == i_)

                particle_X = self.x_center[idx_]
                particle_Y = self.y_center[idx_]
                particle_sigma = self.sigma[idx_]
                particle_labels = self.particle_ID[idx_]

                y = int(particle_Y)
                x = int(particle_X)
                sigma = particle_sigma
                self.imgGrid_list.add_patch(
                    Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))






"""
__author__ = "Houman Mirzaalian D."

https://github.com/mukheshpugal/dictlearn_gpu
"""
from __future__ import print_function

import matplotlib.pyplot as plt

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing.patch_genrator import ImagePatching

from tqdm.autonotebook import tqdm
from sklearn.decomposition import MiniBatchDictionaryLearning, DictionaryLearning
try:
    from dictlearn_gpu import train_dict
    from dictlearn_gpu.utils import dct_dict_1d
    from dictlearn_gpu import OmpBatch
except:
    pass

import numpy as np
import warnings
warnings.filterwarnings('ignore')


class DictionaryLearning_():

    def __init__(self, video, n_comp=100, n_iter=1, non_zero_coeff=10, alpha=0.3, batch_size=3, transform_algorithm='omp',
                 random_select=None, flag_MiniBatchDictionaryLearning=True):
        self.cpu = CPUConfigurations()
        self.tqdm_disable = True
        self.flag_MiniBatchDictionaryLearning = flag_MiniBatchDictionaryLearning

        self.n_comp = n_comp
        self.n_iter = n_iter
        self.non_zero_coeff = non_zero_coeff
        self.alpha = alpha
        self.original_video = video.astype(np.float64)
        self.video_shape = video.shape
        self.batch_size = batch_size
        self.random_select = random_select
        self.transform_algorithm = transform_algorithm
        self.reconstruction_background = None
        self.diff_video = None
        self.patch = None
        self.patch_gen = None
        self.depth, self.width, self.height = None, None, None
        self.depth_overlap, self.width_overlap, self.height_overlap = None, None, None

    def patch_genrator(self, patch_size=16, strides=4):
        self.patch_gen = ImagePatching()
        self.patch = self.patch_gen.split_weight_matrix(self.original_video, patch_size=patch_size, strides=strides)

    def noisy_patches(self, image):
        image = image.astype(np.float64)
        data = image.reshape(image.shape[0], -1)
        mean = np.mean(data, axis=0)
        std = np.std(data, axis=0)
        data -= mean
        data = np.divide(data, std, np.zeros_like(data), where=std != 0)
        return (data, mean, std)

    def ksvd(self, noisy_data):
        if self.flag_MiniBatchDictionaryLearning:
            dic_ = MiniBatchDictionaryLearning(n_components=self.n_comp, alpha=self.alpha, n_iter=self.n_iter,
                                               n_jobs=self.cpu.n_jobs, batch_size=self.batch_size, transform_algorithm=self.transform_algorithm,
                                               transform_n_nonzero_coefs=self.non_zero_coeff)
            dic_.fit(noisy_data)
            return dic_.components_, dic_
        else:
            dic_ = DictionaryLearning(n_components=self.n_comp, alpha=self.alpha, max_iter=self.n_iter, n_jobs=self.cpu.n_jobs)

            dic_.fit(noisy_data)
            return dic_.components_, dic_.error_, dic_

    def ksvd_gpu(self, noisy_data):
        dictionary = dct_dict_1d(n_atoms=self.n_comp, size=noisy_data.shape[0])
        ob = OmpBatch(n_atoms=self.n_comp, sparsity_target=self.non_zero_coeff, batch_size=noisy_data.shape[1])
        decomp = ob.omp_batch(a_0=dictionary.T @ noisy_data, gram=dictionary.T @ dictionary)
        new_dictionary, errors, iters = train_dict(noisy_data, dictionary, sparsity_target=self.non_zero_coeff)
        return new_dictionary, errors, decomp

    def ksvd_kernel(self, i_, dc_correction=False):
        video_patch = self.patch[i_]
        if dc_correction:
            print("---apply dc correction---")
            video_patch = np.divide(video_patch, np.mean(video_patch))

        data_ = video_patch.reshape(video_patch.shape[0], -1)
        if self.random_select is not None:
            num_row = int(self.random_select * data_.shape[0])
            list_selected_idx = np.unique(np.random.randint(0, data_.shape[0], num_row))
            data_selected = data_[list_selected_idx, :]
        if self.flag_MiniBatchDictionaryLearning:
            dict_final, dic = self.ksvd(data_selected)
            code = dic.transform(data_)
            bg_patch_ = np.dot(code, dict_final)
            bg_patch = (bg_patch_.reshape(code.shape[0], video_patch.shape[1], video_patch.shape[2]))
            return bg_patch
        else:
            dict_final, error_iter, dic = self.ksvd(data_selected)
            code = dic.transform(data_)
            bg_patch_ = np.dot(code, dict_final)
            bg_patch = (bg_patch_.reshape(code.shape[0], video_patch.shape[1], video_patch.shape[2]))
            return [bg_patch, error_iter]

    def ksvd_denoising_gpu(self):
        print("\n---apply KSVD on GPU!---")

        if self.patch is not None:
            exp_index_list = []
            pad_non_zero = []
            for i_ in tqdm(range(len(self.patch))):
                video_patch = self.patch[i_]
                data_ = video_patch.reshape(video_patch.shape[0], -1)

                if i_ == 0:
                    big_patch = data_
                else:
                    if data_.shape[1] == big_patch.shape[1]:
                        big_patch = np.concatenate((big_patch, data_), axis=0)
                    else:
                        exp_index_list.append(i_)
                        non_zero = big_patch.shape[1] - data_.shape[1]
                        pad_non_zero.append(non_zero)
                        pad_ = np.zeros((data_.shape[0], non_zero))
                        data_pad = np.concatenate((data_, pad_), axis=1)
                        big_patch = np.concatenate((big_patch, data_pad), axis=0)

            dict_final, dic, decomp = self.ksvd_gpu(big_patch)
            bg_patch_ = dict_final.dot(decomp)

            bg_patch_split = np.split(bg_patch_, len(self.patch), axis=0)

            for pad_, index in zip(pad_non_zero, exp_index_list):
                bg_patch_split[index] = bg_patch_split[index][:, :-pad_]

            cor_patch = []
            for i_, p_ in enumerate(bg_patch_split):
                video_patch = self.patch[i_]
                bg_patch = (p_.reshape(p_.shape[0], video_patch.shape[1], video_patch.shape[2]))
                cor_patch.append(bg_patch)

            self.reconstruction_background = self.patch_gen.reconstruction_weight_matrix(self.video_shape,
                                                                                         new_patch=cor_patch)

            self.diff_video = np.divide(self.original_video, self.reconstruction_background) - 1

            return self.diff_video, self.reconstruction_background

        else:
            print("Patch video is not define!!")

    def ksvd_denoising_cpu(self):
        print("\n---apply KSVD on CPU_new!---")

        if self.patch is not None:
            exp_index_list = []
            pad_non_zero = []
            for i_ in tqdm(range(len(self.patch))):
                video_patch = self.patch[i_]
                data_ = video_patch.reshape(video_patch.shape[0], -1)

                if i_ == 0:
                    big_patch = data_
                else:
                    if data_.shape[1] == big_patch.shape[1]:
                        big_patch = np.concatenate((big_patch, data_), axis=0)
                    else:
                        exp_index_list.append(i_)
                        non_zero = big_patch.shape[1] - data_.shape[1]
                        pad_non_zero.append(non_zero)
                        pad_ = np.zeros((data_.shape[0], non_zero))
                        data_pad = np.concatenate((data_, pad_), axis=1)
                        big_patch = np.concatenate((big_patch, data_pad), axis=0)

            num_row = int(self.random_select * big_patch.shape[0])
            list_selected_idx = np.unique(np.random.randint(0, big_patch.shape[0], num_row))
            data_selected = big_patch[list_selected_idx, :]

            dict_final, dic = self.ksvd(data_selected)
            code = dic.transform(big_patch)
            bg_patch_ = np.dot(code, dict_final)
            # bg_patch = (bg_patch_.reshape(code.shape[0], video_patch.shape[1], video_patch.shape[2]))
            bg_patch_split = np.split(bg_patch_, len(self.patch), axis=0)

            for pad_, index in zip(pad_non_zero, exp_index_list):
                bg_patch_split[index] = bg_patch_split[index][:, :-pad_]

            cor_patch = []
            for i_, p_ in enumerate(bg_patch_split):
                video_patch = self.patch[i_]
                bg_patch = (p_.reshape(p_.shape[0], video_patch.shape[1], video_patch.shape[2]))
                cor_patch.append(bg_patch)

            self.reconstruction_background = self.patch_gen.reconstruction_weight_matrix(self.video_shape,
                                                                                         new_patch=cor_patch)

            self.diff_video = np.divide(self.original_video, self.reconstruction_background) - 1

            return self.diff_video, self.reconstruction_background

        else:
            print("Patch video is not define!!")

    def ksvd_denoising_cpu_(self):
        print("\n---apply KSVD on CPU!---")

        if self.patch is not None:
            bg_patches = [self.ksvd_kernel(x) for x in tqdm(range(len(self.patch)))]

            if self.flag_MiniBatchDictionaryLearning:
                pass
            else:
                temp_ = np.asarray(bg_patches)
                bg_patches = temp_[:, 0]
                error_patchs = temp_[:, 1]
            self.reconstruction_background = self.patch_gen.reconstruction_weight_matrix(self.video_shape,
                                                                                         new_patch=bg_patches)

            self.diff_video = np.divide(self.original_video, self.reconstruction_background) - 1

            return self.diff_video, self.reconstruction_background

        else:
            print("Patch video is not define!!")

    def ksvd_denoising_iter(self, n_iter_min, n_iter_max, stride, dc_correction=False):
        if self.flag_MiniBatchDictionaryLearning:
            mean_error = []
            std_error = []
            n_iter = list(range(n_iter_min, n_iter_max, stride))
            for n_i in tqdm(n_iter, disable=False):
                self.n_iter = n_i
                self.tqdm_disable = True
                if self.patch is not None:
                    bg_patches = [self.ksvd_kernel(x) for x in tqdm(range(len(self.patch)), disable=self.tqdm_disable)]
                    self.reconstruction_background = self.patch_gen.reconstruction_weight_matrix(self.video_shape, new_patch=bg_patches)

                    self.diff_video = np.divide(self.original_video, self.reconstruction_background) - 1

                    diff_ = np.abs(self.original_video - self.reconstruction_background)
                    mean_error.append(np.mean(diff_))
                    std_error.append(np.std(diff_))
                    return self.diff_video, self.reconstruction_background, mean_error, std_error
                else:
                    print("Patch video is not define!!")
        else:
            if self.patch is not None:
                self.n_iter = n_iter_max
                results = [self.ksvd_kernel(x, dc_correction) for x in tqdm(range(len(self.patch)), disable=False)]
                temp_ = np.asarray(results)
                bg_patches = temp_[:, 0]
                error_patchs = temp_[:, 1]
                self.reconstruction_background = self.patch_gen.reconstruction_weight_matrix(self.video_shape, new_patch=bg_patches.tolist())

                self.diff_video = np.divide(self.original_video, self.reconstruction_background) - 1

                return self.diff_video, self.reconstruction_background, error_patchs, None
            else:
                print("Patch video is not define!!")











from piscat.GUI.InputOutput import Reading
from functools import partial

from PySide6 import QtCore
from PySide6 import QtWidgets

import os
import numpy as np


class LoadData_GUI(QtWidgets.QWidget):

    update_output_internal = QtCore.Signal(object)
    update_output_external = QtCore.Signal(object)

    update_tab_index = QtCore.Signal(int)

    def __init__(self):
        super(LoadData_GUI, self).__init__()
        self.original_video = None
        self.filename = None
        self.folder = None
        self.data_in = None

        self.update_output_internal.connect(self.updata_setting)

        self.info_video_setting = None
        self.empty_value_box_flag = False

        self.load = QtWidgets.QPushButton("load video")
        self.load.setAutoDefault(False)
        self.load.clicked.connect(self.read_data)
        self.load.setEnabled(True)
        self.load.setFixedWidth(100)

        self.Next = QtWidgets.QPushButton("Next")
        self.Next.setAutoDefault(False)
        self.Next.clicked.connect(self.update_tab)
        self.Next.setEnabled(True)
        self.Next.setFixedWidth(100)

        self.LineEdit_path = QtWidgets.QTextEdit(self)
        self.LineEdit_path_label = QtWidgets.QLabel("Path:")
        self.LineEdit_path.setFixedHeight(20)
        self.LineEdit_path.setFixedWidth(200)

        self.LineEdit_type = QtWidgets.QLineEdit()
        self.LineEdit_type.setPlaceholderText(".bin")
        self.LineEdit_type_label = QtWidgets.QLabel("Data type:")
        self.LineEdit_type.setFixedHeight(20)
        self.LineEdit_type.setFixedWidth(200)

        self.name_mkdir = QtWidgets.QLineEdit()
        self.name_mkdir.setPlaceholderText("folder_name")
        self.name_mkdir_label = QtWidgets.QLabel("Save folder name:")
        self.name_mkdir.setFixedHeight(20)
        self.name_mkdir.setFixedWidth(200)

        self.LineEdit_img_width = QtWidgets.QTextEdit(self)
        self.LineEdit_img_width_label = QtWidgets.QLabel("img_width:")
        self.LineEdit_img_width.setFixedHeight(20)
        self.LineEdit_img_width.setFixedWidth(200)

        self.LineEdit_img_height = QtWidgets.QTextEdit(self)
        self.LineEdit_img_height_label = QtWidgets.QLabel("img_height:")
        self.LineEdit_img_height.setFixedHeight(20)
        self.LineEdit_img_height.setFixedWidth(200)

        self.LineEdit_img_type = QtWidgets.QTextEdit(self)
        self.LineEdit_img_type_label = QtWidgets.QLabel("image_type:")
        self.LineEdit_img_type.setFixedHeight(20)
        self.LineEdit_img_type.setFixedWidth(200)

        self.LineEdit_frame = QtWidgets.QTextEdit(self)
        self.LineEdit_frame_label = QtWidgets.QLabel("frame:")
        self.LineEdit_frame.setFixedHeight(20)
        self.LineEdit_frame.setFixedWidth(200)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)


        self.setLayout(self.grid)

    def createFirstExclusiveGroup(self):
        groupBox = QtWidgets.QGroupBox("File browser:")

        self.grid_file_browser = QtWidgets.QGridLayout()
        self.grid_file_browser.addWidget(self.load, 0, 0)
        self.grid_file_browser.addWidget(self.LineEdit_path_label, 1, 0)
        self.grid_file_browser.addWidget(self.LineEdit_path, 1, 1)
        self.grid_file_browser.addWidget(self.LineEdit_type_label, 2, 0)
        self.grid_file_browser.addWidget(self.LineEdit_type, 2, 1)

        self.grid_file_browser.addWidget(self.LineEdit_img_width_label, 3, 0)
        self.grid_file_browser.addWidget(self.LineEdit_img_width, 3, 1)
        self.grid_file_browser.addWidget(self.LineEdit_img_height_label, 4, 0)
        self.grid_file_browser.addWidget(self.LineEdit_img_height, 4, 1)
        self.grid_file_browser.addWidget(self.LineEdit_frame_label, 5, 0)
        self.grid_file_browser.addWidget(self.LineEdit_frame, 5, 1)
        self.grid_file_browser.addWidget(self.LineEdit_img_type_label, 6, 0)
        self.grid_file_browser.addWidget(self.LineEdit_img_type, 6, 1)
        self.grid_file_browser.addWidget(self.name_mkdir_label, 7, 0)
        self.grid_file_browser.addWidget(self.name_mkdir, 7, 1)

        self.grid_file_browser.addWidget(self.Next, 8, 0)

        groupBox.setLayout(self.grid_file_browser)
        return groupBox


    def read_data(self):
        reading = Reading()
        reading.update_output.connect(self.updata_input_setting)
        reading.load_batch_data()

    @QtCore.Slot()
    def updata_input_setting(self, data_in):
        self.info_video_setting = data_in[0]

        self.filename = self.info_video_setting['path']
        self.LineEdit_path.setText(self.filename)
        # self.LineEdit_type.setText(self.info_video_setting['title'])

        img_width = self.info_video_setting['img_width']
        img_height = self.info_video_setting['img_height']
        s_frame = self.info_video_setting['s_frame']
        e_frame = self.info_video_setting['e_frame']
        frame_stride = self.info_video_setting['frame_stride']
        width_size_s = self.info_video_setting['width_size_s']
        width_size_e = self.info_video_setting['width_size_e']
        height_size_s = self.info_video_setting['height_size_s']
        height_size_e = self.info_video_setting['height_size_e']
        image_type = str(self.info_video_setting['image_type'])

        imageSize_w = 'img_width:' + str(img_width) + ',' + str(width_size_s) + ':' + str(width_size_e)
        imageSize_h = 'img_height:' + str(img_height) + ',' + str(height_size_s) + ':' + str(height_size_e)
        imageSize_f = str(s_frame) + ':' + str(e_frame) + ':' + str(frame_stride)

        self.LineEdit_img_type.setText(image_type)
        self.LineEdit_frame.setText(imageSize_f)
        self.LineEdit_img_width.setText(imageSize_w)
        self.LineEdit_img_height.setText(imageSize_h)

        self.update_output_internal.emit(data_in[0])

    def get_values(self):
        try:
            self.file_type = self.LineEdit_type.text()
            self.name_mkdir_ = self.name_mkdir.text()
            self.empty_value_box_flag = True

        except:
            self.msg_box3 = QtWidgets.QMessageBox()
            self.msg_box3.setWindowTitle("Warning!")
            self.msg_box3.setText("Please filled all parameters!")
            self.msg_box3.exec_()
            self.empty_value_box_flag = False

    @QtCore.Slot()
    def updata_setting(self, data_in):
        self.data_in = data_in

    @QtCore.Slot()
    def update_tab(self):
        self.get_values()
        if self.empty_value_box_flag:
            self.data_in['type_file'] = self.file_type
            self.data_in['name_mkdir'] = self.name_mkdir_

            self.update_output_external.emit(self.data_in)
            self.update_tab_index.emit(1)







from piscat.GUI.InputOutput import Reading
from functools import partial

from piscat.InputOutput import reading_videos
from piscat.Analysis.analysis_protein_videos import protein_analysis


from PySide6 import QtCore
from PySide6 import QtWidgets

import os
import numpy as np


class ProteinAnalysis_GUI(QtWidgets.QWidget):

    update_output_internal = QtCore.Signal(object)
    update_output_external = QtCore.Signal(object)

    update_tab_index = QtCore.Signal(int)

    def __init__(self):
        super(ProteinAnalysis_GUI, self).__init__()
        self.hyperparameters = None
        self.flags = None

        self.info_video_setting = None
        self.empty_value_box_flag = False

        self.Run = QtWidgets.QPushButton("Run")
        self.Run.setAutoDefault(False)
        self.Run.clicked.connect(self.run_analysis_scr)
        self.Run.setEnabled(True)
        self.Run.setFixedWidth(100)

        self.LineEdit_hyperparameters = QtWidgets.QTextEdit(self)
        self.LineEdit_hyperparameters_label = QtWidgets.QLabel("Hyperparameters:")
        # self.LineEdit_hyperparameters.setFixedHeight(20)
        # self.LineEdit_hyperparameters.setFixedWidth(200)

        self.LineEdit_flags = QtWidgets.QTextEdit(self)
        self.LineEdit_flags_label = QtWidgets.QLabel("flags:")
        # self.LineEdit_flags.setFixedHeight(20)
        # self.LineEdit_flags.setFixedWidth(200)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.createFirstExclusiveGroup(), 0, 0)

        self.setLayout(self.grid)

    def createFirstExclusiveGroup(self):
        groupBox = QtWidgets.QGroupBox("File browser:")

        self.grid_file_browser = QtWidgets.QGridLayout()
        self.grid_file_browser.addWidget(self.LineEdit_hyperparameters_label, 1, 0)
        self.grid_file_browser.addWidget(self.LineEdit_hyperparameters, 1, 1)
        self.grid_file_browser.addWidget(self.LineEdit_flags_label, 2, 0)
        self.grid_file_browser.addWidget(self.LineEdit_flags, 2, 1)
        self.grid_file_browser.addWidget(self.Run, 8, 0)

        groupBox.setLayout(self.grid_file_browser)
        return groupBox

    @QtCore.Slot()
    def updata_input_setting(self, data_in):
        self.hyperparameters = data_in[0]
        self.flags = data_in[1]

        self.LineEdit_hyperparameters.setText(str(self.hyperparameters))
        self.LineEdit_flags.setText(str(self.flags))

    def run_analysis_scr(self):
        if self.hyperparameters is not None and self.flags is not None:
            type_file = self.hyperparameters['type_file']
            dirName = self.hyperparameters['path']
            name_mkdir = self.hyperparameters['name_mkdir']

            df_video = reading_videos.DirectoryType(dirName, type_file).return_df()

            paths = df_video['Directory'].tolist()
            video_names = df_video['File'].tolist()
            protein_analysis(paths=paths, video_names=video_names, hyperparameters=self.hyperparameters, flags=self.flags,
                             name_mkdir=name_mkdir)
        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Hyperparameters or Flags is not defined!")
            self.msg_box.exec_()














from piscat.InputOutput import read_write_data
from piscat.GUI.BatchAnalysis.tab_bg_correction import BgCorrection_GUI
from piscat.GUI.BatchAnalysis.tab_localization import Localization_GUI
from piscat.GUI.BatchAnalysis.tab_reading_data import LoadData_GUI
from piscat.GUI.BatchAnalysis.tab_plot_histogram import Histogram_GUI
from piscat.GUI.BatchAnalysis.tab_trajectory import Tracking_GUI
from piscat.GUI.BatchAnalysis.tab_analysis_protein import ProteinAnalysis_GUI


from piscat.InputOutput.read_write_data import load_dict_from_hdf5

from PySide6 import QtCore, QtWidgets
from functools import partial
import pandas as pd
import numpy as np
import time
import os


class BatchAnalysis(QtWidgets.QMainWindow):

    def __init__(self, object_update_progressBar, parent=None):
        super(BatchAnalysis, self).__init__(parent)
        self.window = QtWidgets.QWidget()

        self.all_tabs = {}
        self.type_bgCorrection = None
        self.current_frame = 0
        self.object_update_progressBar = object_update_progressBar

        self.type_file = None
        self.dirName = None
        self.name_mkdir = None

        self.hyperparameters = {'im_size_x': None, 'im_size_y': None, 'start_fr': None, 'end_fr': None, 'image_format': None,
                                'batch_size': None, 'mode_FPN': 'mFPN', 'select_correction_axis': 'Both',
                                'function': None, 'Mode_PSF_Segmentation': 'BOTH', 'min_sigma': None, 'max_sigma': None, 'sigma_ratio': None, 'PSF_detection_thr': None, 'overlap': None,
                                'search_range': None, 'memory': None,
                                'outlier_frames_thr': None, 'symmetric_PSFs_thr': None, 'min_V_shape_width': None}

        self.flags = {'PN': None, 'DRA': None, 'FPNc': None, 'outlier_frames_filter': None, 'Dense_Filter': None, 'symmetric_PSFs_Filter': None, 'FFT_flag': None}

        self.history = {'Type_bg_correction': None, 'bg_setting': None,
                        'PSFs_Localization_setting': None, 'PSFs_Tracking_setting': None}

        self.PSFs_Particels_num = {'#PSFs': None, '#Particles': None}

        self.initUI()

        self.setWindowTitle("Batch analysis")
        self.window.setLayout(self.layout)
        self.setCentralWidget(self.window)
        self.window.show()

    def initUI(self):
        self.resize(400, 300)
        self.setWindowTitle('Batch analysis')
        self.setGeometry(1140, 100, 400, 200)

        # Tabs
        self.input_video = None
        self.batch_size = None
        self.all_tabs = {"Load_data": LoadData_GUI(),
                         "Bg_correction": BgCorrection_GUI(),
                         "PSFs_Localization": Localization_GUI(),
                         "PSFs_Tracking": Tracking_GUI(self.batch_size),
                         "Protein_Analysis": ProteinAnalysis_GUI(),
                         "Histogram": Histogram_GUI()}

        self.all_tabs["Load_data"].update_output_external.connect(self.update_read_data)
        self.all_tabs["Load_data"].update_tab_index.connect(self.update_tab)

        self.all_tabs["Bg_correction"].output_setting_Tab_bgCorrection.connect(self.update_BGc)
        self.all_tabs["Bg_correction"].update_tab_index.connect(self.update_tab)

        self.all_tabs["PSFs_Localization"].output_setting_Tab_Localization_external.connect(self.Update_tab_localization)
        self.all_tabs["PSFs_Localization"].update_tab_index.connect(self.update_tab)

        self.all_tabs["PSFs_Tracking"].output_setting_Tab_tracking.connect(self.Update_tab_trajectories)
        self.all_tabs["PSFs_Tracking"].update_tab_index.connect(self.update_tab)

        self.main_tabWidget = QtWidgets.QTabWidget()
        self.main_tabWidget.addTab(self.all_tabs["Load_data"], "Load data")
        self.main_tabWidget.addTab(self.all_tabs["Bg_correction"], "Background correction")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Localization"], "PSFs localization")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Tracking"], "PSFs tracking")
        self.main_tabWidget.addTab(self.all_tabs["Protein_Analysis"], "Protein analysis")
        self.main_tabWidget.addTab(self.all_tabs["Histogram"], "Histogram")

        self.btn_save = QtWidgets.QPushButton("Save Data_History")
        self.btn_save.setAutoDefault(False)
        self.btn_save.clicked.connect(self.save_data)
        self.btn_save.setFixedWidth(120)
        self.btn_save.setFixedHeight(20)

        self.btn_load_csv = QtWidgets.QPushButton("Load CSV")
        self.btn_load_csv.setAutoDefault(False)
        self.btn_load_csv.clicked.connect(self.load_csv_data)
        self.btn_load_csv.setFixedWidth(100)
        self.btn_load_csv.setFixedHeight(20)

        self.btn_load_hdf5 = QtWidgets.QPushButton("Load hdf5")
        self.btn_load_hdf5.setAutoDefault(False)
        self.btn_load_hdf5.clicked.connect(self.load_hdf5_data)
        self.btn_load_hdf5.setFixedWidth(120)
        self.btn_load_hdf5.setFixedHeight(20)

        self.statusbar = self.statusBar()
        self.statusBar().addPermanentWidget(self.btn_save)
        self.statusBar().addPermanentWidget(self.btn_load_csv)
        self.statusBar().addPermanentWidget(self.btn_load_hdf5)

        self.main_tabWidget.setCurrentIndex(0)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.main_tabWidget)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def __del__(self):
        print('Destructor called, Employee deleted.')

    @QtCore.Slot(int)
    def update_tab(self, idx_):
        if idx_ == 4:
            self.all_tabs["Protein_Analysis"].updata_input_setting([self.hyperparameters, self.flags])
            self.main_tabWidget.setCurrentIndex(idx_)
        else:
            self.main_tabWidget.setCurrentIndex(idx_)

    @QtCore.Slot(int)
    def update_read_data(self, data_in):
        self.hyperparameters['im_size_x'] = data_in['img_width']
        self.hyperparameters['im_size_y'] = data_in['img_height']
        self.hyperparameters['width_size_s'] = data_in['width_size_s']
        self.hyperparameters['width_size_e'] = data_in['width_size_e']
        self.hyperparameters['height_size_s'] = data_in['height_size_s']
        self.hyperparameters['height_size_e'] = data_in['height_size_e']
        self.hyperparameters['start_fr'] = data_in['s_frame']
        self.hyperparameters['end_fr'] = data_in['e_frame']
        self.hyperparameters['frame_stride'] = data_in['frame_stride']
        self.hyperparameters['image_format'] = data_in['image_type']
        self.hyperparameters['type_file'] = data_in['type_file']
        self.hyperparameters['path'] = data_in['path']
        self.hyperparameters['name_mkdir'] = data_in['name_mkdir']

    @QtCore.Slot()
    def update_BGc(self, data_in):
        for key in data_in:
            if key == 'Batch_size (frames)':
                self.hyperparameters['batch_size'] = data_in[key]
            elif key == 'FPNc_axis':
                self.hyperparameters['select_correction_axis'] = data_in[key]
            elif key == 'Power_Normalization':
                self.flags['PN'] = data_in['Power_Normalization']
            elif key == 'FPNc':
                self.flags[key] = data_in[key]
            elif key == 'filter_hotPixels':
                self.flags[key] = data_in[key]
            elif key == 'type_BGc':
                if data_in['type_BGc'] == 'DRA':
                    self.flags['DRA'] = True
                else:
                    self.flags['DRA'] = False
            else:
                self.hyperparameters[key] = data_in[key]

        self.flags['FFT_flag'] = True

    @QtCore.Slot()
    def Update_tab_localization(self, data_in):

        for key in data_in:
            if key == 'mode':
                self.hyperparameters['Mode_PSF_Segmentation'] = data_in[key]
            elif key == 'threshold_min':
                self.hyperparameters['PSF_detection_thr'] = data_in[key]
            elif key == 'Asymmetry_PSFs_filtering_threshold':
                self.hyperparameters['symmetric_PSFs_thr'] = data_in[key]
            elif key == 'Outlier_frames_filter_max_number_PSFs':
                self.hyperparameters['outlier_frames_thr'] = data_in[key]
            elif key == 'Flag_fine_localization':
                self.flags['Fine_localization'] = data_in[key]
            elif key == 'Flag_fit_2DGaussian':
                self.flags['Flag_fit_2DGaussian'] = data_in[key]
            elif key == 'Side_lobes_Filter' or key == 'Dense_Filter' or key == 'outlier_frames_filter' or key == 'symmetric_PSFs_Filter':
                self.flags[key] = data_in[key]
            else:
                self.hyperparameters[key] = data_in[key]

    @QtCore.Slot()
    def Update_tab_trajectories(self, data_in):
        self.hyperparameters['search_range'] = data_in['Neighborhood_size (px)']
        self.hyperparameters['memory'] = data_in['Memory (frame)']
        self.hyperparameters['min_V_shape_width'] = data_in['Minimum_temporal_length (frame)']

    def save_data(self):
        self.file_path = False
        self.file_path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select Folder', os.path.expanduser("~"), QtWidgets.QFileDialog.ShowDirsOnly)
        timestr = time.strftime("%Y%m%d-%H%M%S")
        name_mkdir = timestr
        try:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            os.mkdir(dr_mk)
            print("Directory ", name_mkdir, " Created ")
        except FileExistsError:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            print("Directory ", name_mkdir, " already exists")

        self.file_path = dr_mk

        if self.file_path:
            if self.df_PSFs is not None:
                read_write_data.save_df2csv(self.df_PSFs, path=self.file_path, name='df_PSFs')

            fout = os.path.join(self.file_path, 'history.txt')
            fo = open(fout, "w")
            for k, v in self.history.items():
                fo.write(str(k) + ' >>> ' + str(v) + '\n\n')

            for k, v in self.PSFs_Particels_num.items():
                fo.write(str(k) + ' >>> ' + str(v) + '\n\n')
            fo.close()

            if self.trajectories is not None:
                read_write_data.save_mat(data=self.trajectories, path=self.file_path, name='all_trajectories')
                read_write_data.save_list_to_hdf5(list_data=self.trajectories, path=self.file_path, name='histData')

    def load_hdf5_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                data_dic = load_dict_from_hdf5(filePath)
                self.Update_tab_trajectories(data_dic)
            except:
                print('HDF5 can not load!')
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()

    def load_csv_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                df_PSFs = pd.read_csv(filePath)
                self.Update_tab_localization(df_PSFs)
            except:
                print('CSV can not load!')
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()




from piscat.Preproccessing.normalization import Normalization
from piscat.BackgroundCorrection.DRA import DifferentialRollingAverage

from piscat.GUI.BackgroundCorrection import DRA
from piscat.GUI.Visualization.fun_display_localization import Visulization_localization
from PySide6 import QtWidgets
from PySide6 import QtCore
from PySide6.QtCore import *


class FUN_DRA(QtWidgets.QMainWindow, QtCore.QObject):

    update_output = QtCore.Signal(object)
    update_pn_roi = QtCore.Signal(object)
    set_new_text = QtCore.Signal(object)
    set_plain_text = QtCore.Signal(object)

    def __init__(self, video, object_update_progressBar):
        super().__init__()

        self.threadpool = QThreadPool()

        self.original_video = video
        self.original_video_pn = None
        self.object_update_progressBar = object_update_progressBar
        self.DRA_video = None
        self.flag_thread_dra = True
        self.flag_update_original_video = True
        self.flag_update_pn_video = False

    @Slot()
    def thread_DRA_complete(self):
        self.object_update_progressBar.setRange(self.p_max)
        self.object_update_progressBar.setLabel('')
        self.thread_DRA.signals.DRA_complete_signal.emit(False)
        self.flag_thread_dra = False
        print("THREAD DRA COMPLETE!")

    @Slot()
    def result_tread_DRA(self, result):
        self.DRA_video = result
        self.flag_update_dra = False
        print("DRA video update!")

    @Slot()
    def update_original_video(self, video):
        if self.flag_update_pn_video:
            self.original_video_pn = video[0]
            self.flag_update_pn_video = False
            self.flag_update_original_video = False

        else:
            self.original_video = video[0]
            self.flag_update_original_video = False

    @Slot()
    def startProgressBar(self, instance, **kwargs):
        self.thread_DRA = instance(**kwargs)
        self.thread_DRA.signals.updateProgress_DRA.connect(self.object_update_progressBar.setProgress)
        self.thread_DRA.signals.result_final.connect(self.result_tread_DRA)
        self.thread_DRA.signals.finished_DRA.connect(self.thread_DRA_complete)

        self.threadpool.start(self.thread_DRA)

    def dra_wrapper(self):
        title = ''

        if self.original_video is not None and self.original_video.shape[0] - (2 * self.info_DRA.batch_size) <= 0:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText(
                "The batch size bigger than the video length (#frame" + str(self.original_video.shape[0]) + ")!")
            self.msg_box.exec_()

        if self.original_video is not None and self.original_video.shape[0] - (2 * self.info_DRA.batch_size) > 0:

            if self.info_DRA.flag_power_normalization:
                self.set_new_text.emit("Start PN -->")

                self.flag_update_original_video = True
                self.flag_update_pn_video = True
                worker = Normalization(self.original_video, flag_pn=True)
                worker.signals.result.connect(self.update_original_video)
                self.threadpool.start(worker)
                while self.flag_update_original_video:
                    QtCore.QCoreApplication.processEvents()

                self.flag_update_original_video = True
                title = title + 'PN_'
                self.set_plain_text.emit(" Done")

            self.p_max = self.original_video.shape[0] - (2 * self.info_DRA.batch_size) - 1
            self.object_update_progressBar.setProgress(0)
            self.object_update_progressBar.setRange(self.p_max)

            if self.info_DRA.flag_FPN:
                self.set_new_text.emit("Start DRA + " + self.info_DRA.mode_FPN + '-->')
                if self.info_DRA.flag_power_normalization:
                    d_arg = {'video': self.original_video_pn, 'batchSize': self.info_DRA.batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag': True, 'select_correction_axis': self.info_DRA.axis,
                             'object_update_progressBar': self.object_update_progressBar, 'mode_FPN': self.info_DRA.mode_FPN}
                else:
                    d_arg = {'video': self.original_video, 'batchSize': self.info_DRA.batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag': True, 'select_correction_axis': self.info_DRA.axis,
                             'object_update_progressBar': self.object_update_progressBar, 'mode_FPN': self.info_DRA.mode_FPN}

                title = title + self.info_DRA.mode_FPN + '_'

            else:
                self.set_new_text.emit("Start DRA -->")
                if self.info_DRA.flag_power_normalization:
                    d_arg = {'video': self.original_video_pn, 'batchSize': self.info_DRA.batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag': False, 'select_correction_axis': self.info_DRA.axis,
                             'object_update_progressBar': self.object_update_progressBar, 'mode_FPN': self.info_DRA.mode_FPN}
                else:

                    d_arg = {'video': self.original_video, 'batchSize': self.info_DRA.batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag': False, 'select_correction_axis': self.info_DRA.axis,
                             'object_update_progressBar': self.object_update_progressBar, 'mode_FPN': self.info_DRA.mode_FPN}

            self.flag_thread_dra = True
            self.flag_update_dra = True
            self.startProgressBar(**d_arg)

            while self.flag_thread_dra:
                QtCore.QCoreApplication.processEvents()

            while self.flag_update_dra:
                QtCore.QCoreApplication.processEvents()

            self.set_plain_text.emit(" Done")

            title = title + 'DRA'

            if self.info_DRA.flag_display:
                self.visualization_ = Visulization_localization()
                self.visualization_.new_display(self.DRA_video, self.DRA_video, object=None, title=title)

            self.update_output.emit([self.DRA_video, title, None, self.info_DRA.batch_size])

        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please load one video!")
            self.msg_box.exec_()

    def apply_DRA(self):
        self.info_DRA = DRA()
        self.info_DRA.signal_finish.connect(self.run_DRA_wrapper)

    def run_DRA_wrapper(self, flag):
        if flag:
            self.info_DRA.window.close()
            self.dra_wrapper()

    def run_DRA_from_bgtabs(self, mode_FPN, batch_size, flag_power_normalization, roi_x_pn, roi_y_pn, flag_FPN, axis):
        title = ''

        if self.original_video is not None and self.original_video.shape[0] - (2 * batch_size) <= 0:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText(
                "The batch size bigger than the video length (#frame" + str(self.original_video.shape[0]) + ")!")
            self.msg_box.exec_()

        if self.original_video is not None and self.original_video.shape[0] - (2 * batch_size) > 0:

            if flag_power_normalization:
                self.set_new_text.emit("Start PN -->")
                self.flag_update_original_video = True
                self.flag_update_pn_video = True

                worker = Normalization(self.original_video, flag_pn=True)
                self.update_pn_roi.connect(worker.update_class_parameter)
                worker.signals.result.connect(self.update_original_video)
                self.update_pn_roi.emit([roi_x_pn, roi_y_pn])
                self.threadpool.start(worker)
                while self.flag_update_original_video:
                    QtCore.QCoreApplication.processEvents()

                self.flag_update_original_video = True
                title = title + 'PN_'
                self.set_plain_text.emit(" Done")

            self.p_max = self.original_video.shape[0] - (2 * batch_size) - 1
            self.object_update_progressBar.setProgress(0)
            self.object_update_progressBar.setRange(self.p_max)

            if flag_FPN:
                self.set_new_text.emit("Start DRA + " + mode_FPN + '-->')
                if flag_power_normalization:
                    d_arg = {'video': self.original_video_pn, 'batchSize': batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag_GUI': True,
                             'gui_select_correction_axis': axis,
                             'object_update_progressBar': self.object_update_progressBar,
                             'mode_FPN': mode_FPN}
                else:
                    d_arg = {'video': self.original_video, 'batchSize': batch_size, 'flag_GUI': True,
                             'instance': DifferentialRollingAverage, 'FPN_flag_GUI': True,
                             'gui_select_correction_axis': axis,
                             'object_update_progressBar': self.object_update_progressBar,
                             'mode_FPN': mode_FPN}

                title = title + mode_FPN + '_'

            else:
                self.set_new_text.emit("Start DRA -->")
                if flag_power_normalization:
                    d_arg = {'video': self.original_video_pn, 'batchSize': batch_size, 'flag_GUI': True, 'mode_FPN': mode_FPN,
                             'instance': DifferentialRollingAverage, 'FPN_flag_GUI': False, 'gui_select_correction_axis': axis,
                             'object_update_progressBar': self.object_update_progressBar}
                else:
                    d_arg = {'video': self.original_video, 'batchSize': batch_size, 'flag_GUI': True, 'mode_FPN': mode_FPN,
                             'instance': DifferentialRollingAverage, 'FPN_flag_GUI': False, 'gui_select_correction_axis': axis,
                             'object_update_progressBar': self.object_update_progressBar}

            self.flag_thread_dra = True
            self.flag_update_dra = True
            self.startProgressBar(**d_arg)

            while self.flag_thread_dra:
                QtCore.QCoreApplication.processEvents()

            while self.flag_update_dra:
                QtCore.QCoreApplication.processEvents()

            self.set_plain_text.emit(" Done")

            title = title + 'DRA'

            return self.DRA_video


        else:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Please load one video!")
            self.msg_box.exec_()




from PySide6 import QtCore, QtWidgets, QtGui

import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Qt5Agg')


class MplCanvas(FigureCanvas):

    def __init__(self, parent=None, width=5, height=4, dpi=100, numRows=1, numColumns=1):
        fig = Figure(figsize=(width, height), dpi=dpi)
        grid = plt.GridSpec(numRows, numColumns, hspace=0.3, wspace=0.3)

        self.axes = []
        for i in range(numRows):
            for j in range(numColumns):
                self.axes.append(fig.add_subplot(grid[i, j]))

        super(MplCanvas, self).__init__(fig)


class UpdatingPlots(QtWidgets.QMainWindow):

    def __init__(self, *args, **kwargs):
        super(UpdatingPlots, self).__init__(*args, **kwargs)

        self.canvas = MplCanvas(self, width=6, height=6, dpi=100)
        self.setCentralWidget(self.canvas)

    def update_plot(self, ydata):
        self.canvas.axes.clear()  # Clear the canvas.
        self.canvas.axes.plot(ydata, 'r.')
        self.canvas.axes.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        self.canvas.axes.grid()
        self.canvas.axes.set_ylabel('Pixel intensity')
        self.canvas.draw()


class UpdatingPlotsPyqtGraphSpatial(QtWidgets.QWidget):

    def __init__(self):
        super(UpdatingPlotsPyqtGraphSpatial, self).__init__()
        self.layout = QtWidgets.QVBoxLayout(self)

        self.graphWidget = pg.PlotWidget()
        self.graphWidget.setBackground('w')
        self.graphWidget.showGrid(x=True, y=True)
        styles = {"color": "#f00", "font-size": "20px"}
        self.graphWidget.setLabel("left", "Pixel intensity", **styles)

        self.layout.addWidget(self.graphWidget)

        pen = pg.mkPen(color=(255, 102, 0))
        self.data_line = self.graphWidget.plot([1], [1], pen=pen)

    def closeEvent(self, event):
        # QtCore.QCoreApplication.instance().quit()
        print("closing plot")


class UpdatingPlotsPyqtGraphTemporal(QtWidgets.QWidget):

    def __init__(self):
        super(UpdatingPlotsPyqtGraphTemporal, self).__init__()
        self.layout = QtWidgets.QVBoxLayout(self)

        self.graphWidget = pg.PlotWidget()
        self.graphWidget.setBackground('w')
        self.graphWidget.showGrid(x=True, y=True)
        styles = {"color": "#f00", "font-size": "20px"}
        self.graphWidget.setLabel("left", "Pixel intensity", **styles)
        self.graphWidget.setLabel("bottom", "Frame number", **styles)

        self.layout.addWidget(self.graphWidget)

        pen_1 = pg.mkPen(color=(255, 102, 0))
        pen_2 = pg.mkPen(color=(0, 0, 255))

        self.data_line = self.graphWidget.plot([1], [1], pen=pen_1)
        self.data_point = self.graphWidget.plot([1], [1], pen=pen_2, symbol='o')

    def closeEvent(self, event):
        print("closing plot")


# class UpdatingPlots_Image(QtWidgets.QMainWindow):
#
#     def __init__(self, list_img, list_titles, x_axis_labels, y_axis_labels, *args, **kwargs):
#         super(UpdatingPlots_Image, self).__init__(*args, **kwargs)
#         self.list_titles = list_titles
#         self.list_img = list_img
#         self.x_axis_labels = x_axis_labels
#         self.y_axis_labels = y_axis_labels
#
#         self.canvas = MplCanvas(self, width=9, height=4, dpi=100, numRows=1, numColumns=3)
#         self.setCentralWidget(self.canvas)
#
#     def update_plot(self, xdata, ydata):
#         for i_ in range(3):
#             self.canvas.axes[i_].clear()  # Clear the canvas.
#             self.canvas.axes[i_].imshow(self.list_img[i_])
#             self.canvas.axes[i_].plot(xdata, ydata, 'r-', alpha=0.3)
#             self.canvas.axes[i_].ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
#             self.canvas.axes[i_].set_ylabel(self.x_axis_labels[i_])
#             self.canvas.axes[i_].set_xlabel(self.y_axis_labels[i_])
#
#         self.canvas.draw()


class UpdatingPlots_Image(QtGui.QMainWindow):

    def __init__(self, list_img, list_titles, x_axis_labels, y_axis_labels, title, *args, **kwargs):
        super(UpdatingPlots_Image, self).__init__(*args, **kwargs)
        self.list_titles = list_titles
        self.list_img = list_img
        self.x_axis_labels = x_axis_labels
        self.y_axis_labels = y_axis_labels

        self.mainbox = QtGui.QWidget()
        self.setCentralWidget(self.mainbox)

        self.canvas = pg.GraphicsLayoutWidget(size=(100, 100), title=title)
        self.label = QtGui.QLabel()

        lay = QtGui.QVBoxLayout(self.mainbox)
        lay.addWidget(self.canvas)
        lay.addWidget(self.label)

        pen = pg.mkPen(color=(255, 0, 0), width=2)

        self.img_items = []
        for i in range(3):
            view = self.canvas.addPlot(0, i)
            view.setAspectLocked(False)
            view.setTitle(self.list_titles[i])
            view.setLabel(axis='left', text=self.x_axis_labels[i])
            view.setLabel(axis='bottom', text=self.y_axis_labels[i])

            view.setRange(QtCore.QRectF(0, 0, self.list_img[i].shape[1], self.list_img[i].shape[0]))
            it = pg.ImageItem(None, border="w")
            it_line = pg.InfiniteLine(pos=None, angle=0, pen=pen, movable=False)
            view.addItem(it)
            view.addItem(it_line)
            self.img_items.append([it, it_line])
            self.canvas.nextRow()

    def update_plot(self, xdata, ydata):
        for i_, item in enumerate(self.img_items):
            item[0].setImage(np.transpose(self.list_img[i_]))
            item[1].setValue(ydata[0])








import os
import sys
import numpy as np

def fixed_length(in_list):
    """
    This function create the list with same length for all sub list.

    Parameters
    ----------
    in_list: list
        List contains some sublist.

    Returns
    -------
    tmp: list
    """
    tmp = np.empty([len(in_list), len(max(in_list, key=lambda x: len(x)))])
    for i, j in enumerate(in_list):
        tmp[i][0:len(j)] = j
    return tmp

def protein_trajectories_list2dic(v_shape_list):
    """
    From the output of ``TemporalFilter.v_trajectory``, this function converts the list to dictionary format.

    Parameters
    ----------
    v_shape_list: List of list

        | [intensity_horizontal, intensity_vertical, particle_center_intensity,
            particle_center_intensity_follow, particle_frame, particle_sigma, particle_X, particle_Y, particle_ID,
            optional(fit_intensity, fit_x, fit_y, fit_X_sigma, fit_Y_sigma, fit_Bias, fit_intensity_error,
            fit_x_error, fit_y_error, fit_X_sigma_error, fit_Y_sigma_error, fit_Bias_error)]

    Returns
    -------
    dic_all: dic
       Return dictionary similar to the following structures

       | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ..., 'particle_ID': ...},
            "#1": {}, ...}

    """
    dic_all = {}

    if type(v_shape_list) is list:
        v_shape_list = np.asarray(v_shape_list)

    if type(v_shape_list) is np.ndarray:

        for n_particle in range(v_shape_list.shape[0]):
            dic_particles = {}

            if type(v_shape_list[n_particle][0]) is list:
                dic_particles['intensity_horizontal'] = np.asarray(fixed_length(v_shape_list[n_particle][0]))
                dic_particles['intensity_vertical'] = np.asarray(fixed_length(v_shape_list[n_particle][1]))
                dic_particles['center_int'] = np.asarray(v_shape_list[n_particle][2])
                dic_particles['center_int_flow'] = np.asarray(v_shape_list[n_particle][3])
                dic_particles['frame_number'] = np.asarray(v_shape_list[n_particle][4])
                dic_particles['sigma'] = np.asarray(v_shape_list[n_particle][5])
                dic_particles['x_center'] = np.asarray(v_shape_list[n_particle][6])
                dic_particles['y_center'] = np.asarray(v_shape_list[n_particle][7])
                dic_particles['particle_ID'] = np.asarray(v_shape_list[n_particle][8])
            else:
                dic_particles['intensity_horizontal'] = v_shape_list[n_particle][0].ravel()
                dic_particles['intensity_vertical'] = v_shape_list[n_particle][1].ravel()
                dic_particles['center_int'] = v_shape_list[n_particle][2].ravel()
                dic_particles['center_int_flow'] = v_shape_list[n_particle][3].ravel()
                dic_particles['frame_number'] = v_shape_list[n_particle][4].ravel()
                dic_particles['sigma'] = v_shape_list[n_particle][5].ravel()
                dic_particles['x_center'] = v_shape_list[n_particle][6].ravel()
                dic_particles['y_center'] = v_shape_list[n_particle][7].ravel()
                dic_particles['particle_ID'] = v_shape_list[n_particle][8].ravel()

            num_parameters = len(v_shape_list[n_particle])

            if num_parameters == 21:
                if type(v_shape_list[n_particle][0]) is list:
                    dic_particles['fit_intensity'] = np.asarray(v_shape_list[n_particle][9])
                    dic_particles['fit_x'] = np.asarray(v_shape_list[n_particle][10])
                    dic_particles['fit_y'] = np.asarray(v_shape_list[n_particle][11])
                    dic_particles['fit_X_sigma'] = np.asarray(v_shape_list[n_particle][12])
                    dic_particles['fit_Y_sigma'] = np.asarray(v_shape_list[n_particle][13])
                    dic_particles['fit_Bias'] = np.asarray(v_shape_list[n_particle][14])
                    dic_particles['fit_intensity_error'] = np.asarray(v_shape_list[n_particle][15])
                    dic_particles['fit_x_error'] = np.asarray(v_shape_list[n_particle][16])
                    dic_particles['fit_y_error'] = np.asarray(v_shape_list[n_particle][17])
                    dic_particles['fit_X_sigma_error'] = np.asarray(v_shape_list[n_particle][18])
                    dic_particles['fit_Y_sigma_error'] = np.asarray(v_shape_list[n_particle][19])
                    dic_particles['fit_Bias_error'] = np.asarray(v_shape_list[n_particle][20])

                else:
                    dic_particles['fit_intensity'] = v_shape_list[n_particle][9].ravel()
                    dic_particles['fit_x'] = v_shape_list[n_particle][10].ravel()
                    dic_particles['fit_y'] = v_shape_list[n_particle][11].ravel()
                    dic_particles['fit_X_sigma'] = v_shape_list[n_particle][12].ravel()
                    dic_particles['fit_Y_sigma'] = v_shape_list[n_particle][13].ravel()
                    dic_particles['fit_Bias'] = v_shape_list[n_particle][14].ravel()
                    dic_particles['fit_intensity_error'] = v_shape_list[n_particle][15].ravel()
                    dic_particles['fit_x_error'] = v_shape_list[n_particle][16].ravel()
                    dic_particles['fit_y_error'] = v_shape_list[n_particle][17].ravel()
                    dic_particles['fit_X_sigma_error'] = v_shape_list[n_particle][18].ravel()
                    dic_particles['fit_Y_sigma_error'] = v_shape_list[n_particle][19].ravel()
                    dic_particles['fit_Bias_error'] = v_shape_list[n_particle][20].ravel()
            dic_all['#'+str(n_particle)] = dic_particles
        return dic_all





import os
import time
import cv2
import imageio
from tqdm.autonotebook import tqdm


def write_binary(dir_path, file_name, data, type='original'):
    """
    This function writes video as a binary.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file.

    data: NDArray
        Video with numpy format.

    type: str or bin_type
        The video bin type is not changed by 'original,' but the user can convert it (e.g. float --> int16).

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    if type == 'original':
        data = data
    else:
        data = data.astype(type)

    data = data.copy(order='C')
    with open(save_path, 'wb') as outfile:
        outfile.write(data)

    return save_path_


def write_MP4(dir_path, file_name, data, jump=0, fps=10):
    """
    This function writes video as a MP4.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file.

    data: NDArray
        Video with numpy format.

    jump: int
        Define stride between frames.

    fps: int
        Number of frame per seconds.

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    data = data.copy(order='C')
    image_ = []
    for frame_number in tqdm(range(0, data.shape[0] - jump, jump)):
        image_.append(data[frame_number, ...])

    imageio.mimsave(save_path, image_, format='MP4', fps=fps)

    return save_path_


def write_GIF(dir_path, file_name, data, jump=0, fps=10):
    """
    This function writes video as a GIF.

    Parameters
    ----------
    dir_path: str
        Path of the directory that video save on it.

    file_name: str
       Name of the save file

    data: NDArray
        Video with numpy format.

    jump: int
        Define stride between frames.

    fps: int
        Number of frame per seconds.

    Returns
    -------
    The path to the new folder that was created to save the video is returned.
    """
    timestr = time.strftime("%Y%m%d-%H%M%S")
    try:
        dr_mk = os.path.join(dir_path, timestr)
        os.mkdir(dr_mk)
        print("Directory ", timestr, " Created ")
    except FileExistsError:
        print("Directory ", timestr, " already exists")

    save_path = os.path.join(dir_path, timestr, file_name)
    save_path_ = os.path.join(dir_path, timestr)

    data = data.copy(order='C')
    image_ = []
    for frame_number in tqdm(range(0, data.shape[0] - jump, jump)):
        image_.append(data[frame_number, ...])

    imageio.mimsave(save_path, image_, format='GIF', fps=fps)
    return save_path_



import numpy as np
from joblib import Parallel, delayed
from PySide6 import QtGui
from PySide6 import QtCore
from PySide6 import QtWidgets
from PySide6.QtCore import *
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from tqdm.autonotebook import tqdm


class WorkerSignals(QObject):

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class Normalization(QRunnable):

    def __init__(self, video, flag_pn=False, flag_global=False, flag_image_specific=False):
        """
        This class contains a different version of video/image normalization methods.

        Parameters
        ----------
        video: NDArray
            It is Numpy 3D video array.

        optional_1: GUI
            * `flag_pn`: bool
                If it is True, the power_normalized method is applied on input video on GUI.

            * `flag_global`: bool
                If it is True, the normalized_image_global method is applied on input video on GUI.

            * `flag_image_specific`: bool
                If it is True, the normalized_image_specific method is applied on input video on GUI.

        """
        super(Normalization, self).__init__()
        self.cpu = CPUConfigurations()

        self.video = video
        self.roi_x = None
        self.roi_y = None
        self.signals = WorkerSignals()

        self.flag_pn = flag_pn
        self.flag_global = flag_global
        self.flag_image_specific = flag_image_specific

    @Slot()
    def update_class_parameter(self, data_in):
        self.roi_x = data_in[0]
        self.roi_y = data_in[1]

    @Slot()
    def run(self, *args, **kwargs):
        if self.flag_pn:
            result = self.power_normalized(self.roi_x, self.roi_y)
        if self.flag_global:
            result = self.normalized_image_global()
        if self.flag_image_specific:
            result = self.normalized_image_specific()

        self.signals.result.emit(result)

    def normalized_image_global(self, new_max=1, new_min=0):
        """
        Based on the global min and max in the video, this method normalizes all pixels in the video between``new_min`` and ``new_max``.

        Parameters
        ----------
        new_max: float
            Video's new global maximum pixel intensity.

        new_min: float
            Video's new global minimum pixel intensity.

        Returns
        -------
        img2: NDArray
            Normalize video (3D-Numpy array).
        """
        mins = self.video.min(axis=(0, 1, 2), keepdims=True)
        maxs = self.video.max(axis=(0, 1, 2), keepdims=True)
        img2 = ((self.video - mins) * (new_max - new_min)) / (maxs - mins) + new_min
        return img2

    def normalized_image_specific(self, scale=255, format='uint8'):
        """
        This approach normalizes all pixels in the image between 0 and 1 based on the image min and max in the video frame.

        Parameters
        ----------
        scale: float
            Video's new global maximum pixel intensity.

        format: str
            It describes how the bytes in the fixed-size block of memory corresponding to an array item should be interpreted.

        Returns
        -------
        n_video: NDArray
            Normalize video (3D-Numpy array).
        """
        if len(self.video.shape) == 3:
            print('\nconverting video bin_type to ' + format + '--->', end=" ")

            tmp_0 = np.reshape(self.video, (self.video.shape[0], -1))

            min_tmp1 = np.expand_dims(tmp_0.min(axis=1), axis=1)

            ptp_tmp1 = np.expand_dims(tmp_0.ptp(axis=1), axis=1)

            tmp2 = np.subtract(tmp_0, min_tmp1)
            tmp3 = np.divide(tmp2, ptp_tmp1)
            n_video_ = scale * np.reshape(tmp3, (self.video.shape[0], self.video.shape[1], self.video.shape[2]))
            n_video = n_video_.astype(format)
            print('Done')

        elif len(self.video.shape) == 2:

            min_tmp1 = self.video.min()

            ptp_tmp1 = self.video.ptp()

            tmp2 = np.subtract(self.video, min_tmp1)
            tmp3 = np.divide(tmp2, ptp_tmp1)
            n_video_ = scale * tmp3
            n_video = n_video_.astype(format)

        return n_video

    def normalized_image_specific_by_max(self):
        print('\nnormalize image by max --->', end=" ")
        n_video = np.empty_like(self.video, dtype=np.float64)
        for i in range(self.video.shape[0]):
            img = self.video[i, :, :]
            n_video[i, :, :] = np.divide(img, img.max())
        print('Done')
        return n_video

    def normalized_image(self):
        """
        Normalization of video between 0 and 1.

        Returns
        -------
        n_video: NDArray
            Normalize video (3D-Numpy array).
        """
        mins = self.video.min()
        maxs = self.video.max()
        img2 = (self.video - mins) / (maxs - mins)
        return img2

    def power_normalized(self, roi_x=None, roi_y=None, inter_flag_parallel_active=False):
        """
        This function corrects the fluctuations in the laser light intensity
        by dividing each pixel in an image by the sum of all pixels on the same frames.

        Parameters
        ----------
        roi_x: tuple
            On the x-axis, is a tuple of the region's minimum and maximum values.

        roi_y: tuple
            On the y-axis, is a tuple of the region's minimum and maximum values.

        inter_flag_parallel_active: bool
            Internal flag for activating parallel computation. Default is False!

        Returns
        -------
        normalized_power: NDArray
            Normalize video (3D-Numpy array).

        power_fluctuation_percentage: NDArray
            Temporal fluctuations of all pixels after power normalization.
        """
        if roi_x is not None or roi_y is not None:
            roi_ = {'x_min': roi_x[0], 'x_max': roi_x[1], 'y_min': roi_y[0], 'y_max': roi_y[1]}
            temp0 = np.sum(self.video[:, roi_['x_min']:roi_['x_max'], roi_['y_min']:roi_['y_max']], axis=1)

        else:
            roi_ = None
            temp0 = np.sum(self.video, axis=1)

        sum_pixels = np.sum(temp0, axis=1)
        if inter_flag_parallel_active is True and self.cpu.parallel_active is True:
            print("\n---start power_normalized with parallel loop---")
            result = Parallel(n_jobs=self.cpu.n_jobs, backend="threading", verbose=self.cpu.verbose)(
                delayed(self.power_normalized_kernel)(f_, roi_) for f_ in tqdm(range(self.video.shape[0])))
            normalized_power = np.asarray(result)
            normalized_power = normalized_power * np.mean(sum_pixels)

            power_fluctuation_percentage = np.divide(sum_pixels, np.mean(sum_pixels)) - 1

        else:
            print("\nstart power_normalized without parallel loop--->", end=" ")
            tmp_ = np.repeat(sum_pixels, self.video.shape[1] * self.video.shape[2])
            temp2 = np.reshape(tmp_, (self.video.shape[0], self.video.shape[1], self.video.shape[2]))
            normalized_power = np.divide(self.video, temp2) * np.mean(sum_pixels)
            power_fluctuation_percentage = np.divide(sum_pixels, np.mean(sum_pixels)) - 1
            print("Done")
        return (normalized_power, power_fluctuation_percentage)

    def power_normalized_kernel(self, f_, roi_):
        if roi_ is not None:
            sum_img_pixels = np.sum(self.video[f_, roi_['x_min']:roi_['x_max'],  roi_['y_min']:roi_['y_max']])
        return np.divide(self.video[f_], sum_img_pixels)


from piscat.InputOutput.cpu_configurations import CPUConfigurations
from joblib import Parallel, delayed

import numpy as np
import cv2 as cv
from tqdm.autonotebook import tqdm
import matplotlib as mpl
import matplotlib.pylab as plt


class Mask2Video:
    def __init__(self, video, mask, inter_flag_parallel_active=True):
        """
        This class produces a masked version of the input video.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        mask: NDArray
            The mask is 2D-numpy with binary values and the same dimension as the input video.

         internal_parallel_flag: bool
            Internal flag for activating parallel computation. Default is True!
        """
        
        self.video = video
        self.cpu = CPUConfigurations()
        self.inter_flag_parallel_active = inter_flag_parallel_active
        if mask is not None:
            if mask.ndim == 2:
                print('--- Same mask is used for all frame! ---')
                self.mask_repeater(mask)

            elif video.shape == mask.shape:
                print('--- Mask has same shape with video! ---')
                self.mask = mask
            else:
                print('--- Mask does not have same shape with video! ---')
        else:
            print('--- Mask is not define! ---')

    def apply_mask(self, flag_nan=True):
        """
        This method is used to apply a mask on a video. The masked values might be defined as nan, or they can be chosen from the median values of each frame.

        Parameters
        ----------
        flag_nan: bool
           A flag that defined how masked values are chosen.

        Returns
        -------
        mask_video_: NDArray
            The video after applying the mask.

        """
        mask_video = []
        if self.inter_flag_parallel_active and self.cpu.parallel_active:
            print("\n---apply mask with Parallel---")
            mask_video = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(
                self.apply_mask_kernel)(f_) for f_ in tqdm(range(self.video.shape[0])))
            mask_video_ = np.asarray(mask_video)

        else:
            print("\n---apply mask without Parallel---")
            for f_ in tqdm(range(self.video.shape[0])):
                img_bg = cv.bitwise_and(self.video[f_], self.video[f_], mask=self.mask[f_])
                mask_video.append(img_bg)
            mask_video_ = np.asarray(mask_video)
        if flag_nan:
            mask_video_[mask_video_ == 0] = np.nan
        else:
            mask_video_ = mask_video_.astype('float64')
            mask_video_[mask_video_ == 0.] = np.nan

            for f_ in tqdm(range(mask_video_.shape[0])):
                img_mask = mask_video_[f_, ...]
                tmp_median = np.nanmedian(img_mask)
                inds = np.where(np.isnan(img_mask))
                img_mask[inds] = tmp_median
                mask_video_[f_, ...] = img_mask
        return mask_video_

    def apply_mask_kernel(self, f_):
        return cv.bitwise_and(self.video[f_], self.video[f_], mask=self.mask[f_])

    def mask_repeater(self, mask):
        """
        Making a 3D numpy array from a 2D mask.

        Parameters
        ----------
        mask: NDArray
            The mask is 2D-numpy with binary values and the same dimension as the input video.
        """
        mask = np.expand_dims(mask, axis=0)
        self.mask = np.repeat(mask, self.video.shape[0], axis=0)

    def mask_generating_circle(self, center=(10, 10), redius=1):
        """
        This method generates a circular mask depending on the rediuse and origin parameters.

        Parameters
        ----------
        center: list
            A list of two integer numbers indicating the mask's center.

        redius: int
            Circular mask redius.

        Returns
        -------
            The mask that has the same shape as the input video.
        """
        mask = np.zeros(self.video[0].shape, dtype="uint8")
        cv.circle(mask, center, redius, 1, -1)
        self.mask_repeater(mask)
        return self.mask


